# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from typing import Tuple

from .basewake import BaseWake, _handle_kind
from .wit import ComponentResonator


class WakeResonator(BaseWake):

    def __init__(self, kind: str = None,
                plane: str = None,
                source_exponents: Tuple[int, int] | None = None,
                test_exponents: Tuple[int, int] | None = None,
                r: float = None, q: float = None, f_r: float = None,
                f_roi_level: float = 0.5):

        if kind is not None:

            kind = _handle_kind(kind)

            components = []
            for kk in kind.keys():
                ff = kind[kk]
                if ff == 0:
                    continue
                cc = ComponentResonator(r=r, q=q, f_r=f_r,
                                        kind=kk, f_roi_level=f_roi_level,
                                        factor=ff, plane=plane,)
                components.append(cc)
        else:
            cc = ComponentResonator(r=r, q=q, f_r=f_r, f_roi_level=f_roi_level,
                                    plane=plane, source_exponents=source_exponents,
                                    test_exponents=test_exponents)
            components = [cc]

        self.components = components
        self.kind = kind



# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from ._version import __version__
from .general import _pkg_root

from .init_pywit_directory import initialize_pywit_directory
from .resonator import WakeResonator
from .wakefield_from_table import WakeFromTable
from .thick_resistive_wall import WakeThickResistiveWall
from .yokoya import Yokoya
from .read_headtail_table import read_headtail_file
from .config_pipeline_for_wakes import config_pipeline_for_wakes
from .beam_elements.transverse_damper import TransverseDamper
from .beam_elements.collective_monitor import CollectiveMonitor




# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import xwakes as xw
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

import json
import h5py

import numpy as np


@for_all_test_contexts
def test_bunch_monitor_hdf5(test_context):
    n_macroparticles = int(1e6)
    zeta_range = (-1, 1)

    energy = 7e12

    offs_x = 1e-3
    offs_px = 2e-3
    offs_y = 3e-3
    offs_py = 4e-3
    offs_zeta = 5e-3
    offs_delta = 6e-3

    sigma_x = 1e-3
    sigma_px = 2e-3
    sigma_y = 3e-3
    sigma_py = 4e-3
    sigma_zeta = 5e-3
    sigma_delta = 6e-3

    # base coordinates
    x_coord = sigma_x * np.random.random(n_macroparticles) + offs_x
    px_coord = sigma_px * np.random.random(n_macroparticles) + offs_px
    y_coord = sigma_y * np.random.random(n_macroparticles) + offs_y
    py_coord = sigma_py * np.random.random(n_macroparticles) + offs_py
    zeta_coord = sigma_zeta * np.random.random(n_macroparticles) + offs_zeta
    delta_coord = sigma_delta * np.random.random(n_macroparticles) + offs_delta

    bunch_spacing_zeta = 10

    n_bunches = 3

    # n_bunches bunches with n_macroparticles each with different coordinates
    particles = xt.Particles(
        _context=test_context, p0c=energy,
        x=np.concatenate([x_coord * (bid + 1) for bid in range(n_bunches)]),
        px=np.concatenate([px_coord * (bid + 1) for bid in range(n_bunches)]),
        y=np.concatenate([y_coord * (bid + 1) for bid in range(n_bunches)]),
        py=np.concatenate([py_coord * (bid + 1) for bid in range(n_bunches)]),
        zeta=np.concatenate([zeta_coord * (bid + 1) - bunch_spacing_zeta * bid
                             for bid in range(n_bunches)]),
        delta=np.concatenate(
            [delta_coord * (bid + 1) for bid in range(n_bunches)]),
    )

    beta = particles.beta0[0]
    gamma = np.sqrt(1 / (1 - beta ** 2))

    # dummy filling scheme
    filling_scheme = np.ones(n_bunches, dtype=int)
    bunch_numbers = np.arange(n_bunches, dtype=int)

    flush_data_every = 10

    monitor = xw.CollectiveMonitor(
        base_file_name='test_monitor',
        backend='hdf5',
        monitor_bunches=True,
        monitor_slices=False,
        monitor_particles=False,
        flush_data_every=flush_data_every,
        zeta_range=zeta_range,
        filling_scheme=filling_scheme,
        bunch_spacing_zeta=bunch_spacing_zeta,
    )

    # track for twice flush_data_every turns so that we test the reshaping
    for _ in range(2 * flush_data_every):
        monitor.track(particles)

    with h5py.File(monitor._bunches_filename, 'r') as h5file:
        for bid in bunch_numbers:
            bunch = h5file[str(bid)]
            print(bunch['mean_x'][:], np.mean(x_coord) * (bid + 1))
            assert np.allclose(bunch['mean_x'][:],
                               np.mean(x_coord) * (bid + 1))
            assert np.allclose(bunch['mean_px'][:],
                               np.mean(px_coord) * (bid + 1))
            assert np.allclose(bunch['mean_y'][:],
                               np.mean(y_coord) * (bid + 1))
            assert np.allclose(bunch['mean_py'][:],
                               np.mean(py_coord) * (bid + 1))
            assert np.allclose(bunch['mean_zeta'][:],
                               (np.mean(zeta_coord * (bid + 1) -
                                        bunch_spacing_zeta * bid)))
            assert np.allclose(bunch['mean_delta'][:],
                               np.mean(delta_coord) * (bid + 1))

            assert np.allclose(bunch['sigma_x'][:],
                               np.std(x_coord) * (bid + 1))
            assert np.allclose(bunch['sigma_px'][:],
                               np.std(px_coord) * (bid + 1))
            assert np.allclose(bunch['sigma_y'][:],
                               np.std(y_coord) * (bid + 1))
            assert np.allclose(bunch['sigma_py'][:],
                               np.std(py_coord) * (bid + 1))
            assert np.allclose(bunch['sigma_zeta'][:],
                               np.std(zeta_coord) * (bid + 1))
            assert np.allclose(bunch['sigma_delta'][:],
                               np.std(delta_coord) * (bid + 1))

            epsn_x = np.sqrt(
                np.linalg.det(np.cov(x_coord, px_coord))) * beta * gamma
            epsn_y = np.sqrt(
                np.linalg.det(np.cov(y_coord, py_coord))) * beta * gamma
            epsn_zeta = np.sqrt(
                np.linalg.det(np.cov(zeta_coord, delta_coord))) * beta * gamma

            assert np.allclose(bunch['epsn_x'][:],
                               epsn_x * (bid + 1) ** 2)
            assert np.allclose(bunch['epsn_y'][:],
                               epsn_y * (bid + 1) ** 2)
            assert np.allclose(bunch['epsn_zeta'][:],
                               epsn_zeta * (bid + 1) ** 2)

            assert np.allclose(bunch['num_particles'][:], n_macroparticles)


@for_all_test_contexts
def test_bunch_monitor_json(test_context):
    n_macroparticles = int(1e6)
    zeta_range = (-1, 1)

    energy = 7e12

    offs_x = 1e-3
    offs_px = 2e-3
    offs_y = 3e-3
    offs_py = 4e-3
    offs_zeta = 5e-3
    offs_delta = 6e-3

    sigma_x = 1e-3
    sigma_px = 2e-3
    sigma_y = 3e-3
    sigma_py = 4e-3
    sigma_zeta = 5e-3
    sigma_delta = 6e-3

    # base coordinates
    x_coord = sigma_x * np.random.random(n_macroparticles) + offs_x
    px_coord = sigma_px * np.random.random(n_macroparticles) + offs_px
    y_coord = sigma_y * np.random.random(n_macroparticles) + offs_y
    py_coord = sigma_py * np.random.random(n_macroparticles) + offs_py
    zeta_coord = sigma_zeta * np.random.random(n_macroparticles) + offs_zeta
    delta_coord = sigma_delta * np.random.random(n_macroparticles) + offs_delta

    bunch_spacing_zeta = 10

    n_bunches = 3

    # n_bunches bunches with n_macroparticles each with different coordinates
    particles = xt.Particles(
        _context=test_context, p0c=energy,
        x=np.concatenate([x_coord * (bid + 1) for bid in range(n_bunches)]),
        px=np.concatenate([px_coord * (bid + 1) for bid in range(n_bunches)]),
        y=np.concatenate([y_coord * (bid + 1) for bid in range(n_bunches)]),
        py=np.concatenate([py_coord * (bid + 1) for bid in range(n_bunches)]),
        zeta=np.concatenate([zeta_coord * (bid + 1) - bunch_spacing_zeta * bid
                             for bid in range(n_bunches)]),
        delta=np.concatenate(
            [delta_coord * (bid + 1) for bid in range(n_bunches)]),
    )

    beta = particles.beta0[0]
    gamma = np.sqrt(1 / (1 - beta ** 2))

    # dummy filling scheme
    filling_scheme = np.ones(n_bunches, dtype=int)
    bunch_numbers = np.arange(n_bunches, dtype=int)

    n_turns = 10

    monitor = xw.CollectiveMonitor(
        base_file_name='test_monitor',
        backend='json',
        monitor_bunches=True,
        monitor_slices=False,
        monitor_particles=False,
        flush_data_every=n_turns,
        zeta_range=zeta_range,
        filling_scheme=filling_scheme,
        bunch_spacing_zeta=bunch_spacing_zeta,
    )

    # track for twice flush_data_every turns so that we test the reshaping
    for _ in range(2 * n_turns):
        monitor.track(particles)

    with open(monitor._bunches_filename, 'r') as f:
        bunches_dict = json.load(f)

    for bid in bunch_numbers:
        bunch = bunches_dict[str(bid)]
        assert np.allclose(bunch['mean_x'][:], (np.mean(x_coord) *
                                                (bid + 1)))
        assert np.allclose(bunch['mean_px'][:], (np.mean(px_coord) *
                                                 (bid + 1)))
        assert np.allclose(bunch['mean_y'][:], (np.mean(y_coord) *
                                                (bid + 1)))
        assert np.allclose(bunch['mean_py'][:], (np.mean(py_coord) *
                                                 (bid + 1)))
        assert np.allclose(bunch['mean_zeta'][:],
                           (np.mean(zeta_coord * (bid + 1) -
                                    bunch_spacing_zeta * bid)))
        assert np.allclose(bunch['mean_delta'][:],
                           np.mean(delta_coord) * (bid + 1))

        assert np.allclose(bunch['sigma_x'][:], (np.std(x_coord) *
                                                 (bid + 1)))
        assert np.allclose(bunch['sigma_px'][:], (np.std(px_coord) *
                                                  (bid + 1)))
        assert np.allclose(bunch['sigma_y'][:], (np.std(y_coord) *
                                                 (bid + 1)))
        assert np.allclose(bunch['sigma_py'][:], (np.std(py_coord) *
                                                  (bid + 1)))
        assert np.allclose(bunch['sigma_zeta'][:],
                           (np.std(zeta_coord) *
                            (bid + 1)))
        assert np.allclose(bunch['sigma_delta'][:],
                           (np.std(delta_coord) * (bid + 1)))

        epsn_x = np.sqrt(
            np.linalg.det(np.cov(x_coord, px_coord))) * beta * gamma
        epsn_y = np.sqrt(
            np.linalg.det(np.cov(y_coord, py_coord))) * beta * gamma
        epsn_zeta = np.sqrt(np.linalg.det(np.cov(zeta_coord,
                                                 delta_coord))) * beta * gamma

        assert np.allclose(bunch['epsn_x'][:], epsn_x * (bid + 1) ** 2)
        assert np.allclose(bunch['epsn_y'][:], epsn_y * (bid + 1) ** 2)
        assert np.allclose(bunch['epsn_zeta'][:], epsn_zeta * (bid + 1) ** 2)

        assert np.allclose(bunch['num_particles'][:], n_macroparticles)


@for_all_test_contexts
def test_slice_monitor_hdf5(test_context):
    n_macroparticles = int(1e6)
    num_slices = 10
    zeta_range = (0, 1)

    energy = 7e12

    offs_x = 1
    offs_px = 2
    offs_y = 3
    offs_py = 4
    offs_delta = 5

    sigma_x = 6
    sigma_px = 7
    sigma_y = 8
    sigma_py = 9
    sigma_delta = 10

    # dummy filling scheme
    n_bunches = 3
    filling_scheme = np.ones(n_bunches, dtype=int)
    bunch_numbers = np.arange(n_bunches, dtype=int)
    bunch_spacing_zeta = 10

    flush_data_every = 10

    monitor = xw.CollectiveMonitor(
        base_file_name='test_monitor',
        backend='hdf5',
        monitor_bunches=False,
        monitor_slices=True,
        monitor_particles=False,
        flush_data_every=flush_data_every,
        zeta_range=zeta_range,
        num_slices=num_slices,
        bunch_spacing_zeta=bunch_spacing_zeta,
        filling_scheme=filling_scheme,
    )

    particles = xt.Particles(
        _context=test_context,
        p0c=energy,
        x=np.zeros(n_macroparticles),
    )

    beta = particles.beta0[0]
    gamma = np.sqrt(1 / (1 - beta ** 2))

    n_slice = 5

    dzeta = monitor.slicer.dzeta
    zeta_coord_base = np.random.random(n_macroparticles)

    x_coord_bunch = []
    px_coord_bunch = []
    y_coord_bunch = []
    py_coord_bunch = []
    zeta_coord_bunch = []
    delta_coord_bunch = []
    n_macroparticles_slice_bunch = []

    for bid in range(n_bunches):
        zeta_coord = zeta_coord_base * (bid + 1) - bid * bunch_spacing_zeta

        bin_min = monitor.slicer.zeta_centers[bid, n_slice] - dzeta / 2
        bin_max = monitor.slicer.zeta_centers[bid, n_slice] + dzeta / 2

        slice_mask = np.logical_and(zeta_coord < bin_max, zeta_coord >= bin_min)

        n_macroparticles_slice = np.sum(slice_mask)

        x_coord = (sigma_x * np.random.random(
            n_macroparticles_slice) + offs_x) * (bid + 1)
        px_coord = (sigma_px * np.random.random(
            n_macroparticles_slice) + offs_px) * (bid + 1)
        y_coord = (sigma_y * np.random.random(
            n_macroparticles_slice) + offs_y) * (bid + 1)
        py_coord = (sigma_py * np.random.random(
            n_macroparticles_slice) + offs_py) * (bid + 1)
        delta_coord = (sigma_delta * np.random.random(n_macroparticles_slice) +
                       offs_delta) * (bid + 1)

        x_coord_bunch.append(x_coord)
        px_coord_bunch.append(px_coord)
        y_coord_bunch.append(y_coord)
        py_coord_bunch.append(py_coord)
        delta_coord_bunch.append(delta_coord)
        zeta_coord_bunch.append(zeta_coord[slice_mask])
        n_macroparticles_slice_bunch.append(n_macroparticles_slice)

        particles.x[slice_mask] = x_coord
        particles.px[slice_mask] = px_coord
        particles.y[slice_mask] = y_coord
        particles.py[slice_mask] = py_coord
        particles.zeta[slice_mask] = zeta_coord[slice_mask]
        particles.delta[slice_mask] = delta_coord

    # track for twice flush_data_every turns so that we test the reshaping
    for _ in range(flush_data_every * 2):
        monitor.track(particles)

    with h5py.File(monitor._slices_filename, 'r') as h5file:
        for bid in bunch_numbers:
            slice_data = h5file[str(bid)]
            print(x_coord_bunch[bid])
            assert np.allclose(slice_data['mean_x'][:, n_slice],
                               np.mean(x_coord_bunch[bid]))
            assert np.allclose(slice_data['mean_px'][:, n_slice],
                               np.mean(px_coord_bunch[bid]))
            assert np.allclose(slice_data['mean_y'][:, n_slice],
                               np.mean(y_coord_bunch[bid]))
            assert np.allclose(slice_data['mean_py'][:, n_slice],
                               np.mean(py_coord_bunch[bid]))
            assert np.allclose(slice_data['mean_zeta'][:, n_slice],
                               np.mean(zeta_coord_bunch[bid]))
            assert np.allclose(slice_data['mean_delta'][:, n_slice],
                               np.mean(delta_coord_bunch[bid]))

            assert np.allclose(slice_data['sigma_x'][:, n_slice],
                               np.std(x_coord_bunch[bid]))
            assert np.allclose(slice_data['sigma_px'][:, n_slice],
                               np.std(px_coord_bunch[bid]))
            assert np.allclose(slice_data['sigma_y'][:, n_slice],
                               np.std(y_coord_bunch[bid]))
            assert np.allclose(slice_data['sigma_py'][:, n_slice],
                               np.std(py_coord_bunch[bid]))
            assert np.allclose(slice_data['sigma_zeta'][:, n_slice],
                               np.std(zeta_coord_bunch[bid]))
            assert np.allclose(slice_data['sigma_delta'][:, n_slice],
                               np.std(delta_coord_bunch[bid]))

            epsn_x = (np.sqrt(np.var(x_coord_bunch[bid]) *
                              np.var(px_coord_bunch[bid]) -
                              np.cov(x_coord_bunch[bid], px_coord_bunch[bid])[
                                  0, 1])
                      * beta * gamma)
            epsn_y = (np.sqrt(np.var(y_coord_bunch[bid]) *
                              np.var(py_coord_bunch[bid]) -
                              np.cov(y_coord_bunch[bid], py_coord_bunch[bid])[
                                  0, 1])
                      * beta * gamma)
            epsn_zeta = (np.sqrt(np.var(zeta_coord_bunch[bid]) *
                                 np.var(delta_coord_bunch[bid]) -
                                 np.cov(zeta_coord_bunch[bid],
                                        delta_coord_bunch[bid])[
                                     0, 1]) * beta * gamma)

            assert np.allclose(slice_data['epsn_x'][:, n_slice], epsn_x)
            assert np.allclose(slice_data['epsn_y'][:, n_slice], epsn_y)
            assert np.allclose(slice_data['epsn_zeta'][:, n_slice], epsn_zeta)

            assert np.allclose(slice_data['num_particles'][:, n_slice],
                               n_macroparticles_slice_bunch[bid])


@for_all_test_contexts
def test_slice_monitor_json(test_context):
    n_macroparticles = int(1e6)
    num_slices = 10
    zeta_range = (0, 1)

    energy = 7e12

    offs_x = 1
    offs_px = 2
    offs_y = 3
    offs_py = 4
    offs_delta = 5

    sigma_x = 6
    sigma_px = 7
    sigma_y = 8
    sigma_py = 9
    sigma_delta = 10

    # dummy filling scheme
    n_bunches = 3
    filling_scheme = np.ones(n_bunches, dtype=int)
    bunch_numbers = np.arange(n_bunches, dtype=int)
    bunch_spacing_zeta = 10

    flush_data_every = 10

    monitor = xw.CollectiveMonitor(
        base_file_name='test_monitor',
        backend='json',
        monitor_bunches=False,
        monitor_slices=True,
        monitor_particles=False,
        flush_data_every=flush_data_every,
        zeta_range=zeta_range,
        num_slices=num_slices,
        bunch_spacing_zeta=bunch_spacing_zeta,
        filling_scheme=filling_scheme,
    )

    particles = xt.Particles(
        _context=test_context,
        p0c=energy,
        x=np.zeros(n_macroparticles),
    )

    beta = particles.beta0[0]
    gamma = np.sqrt(1 / (1 - beta ** 2))

    n_slice = 5

    dzeta = monitor.slicer.dzeta
    zeta_coord_base = np.random.random(n_macroparticles)

    x_coord_bunch = []
    px_coord_bunch = []
    y_coord_bunch = []
    py_coord_bunch = []
    zeta_coord_bunch = []
    delta_coord_bunch = []
    n_macroparticles_slice_bunch = []

    for bid in range(n_bunches):
        zeta_coord = zeta_coord_base * (bid + 1) - bid * bunch_spacing_zeta

        bin_min = monitor.slicer.zeta_centers[bid, n_slice] - dzeta / 2
        bin_max = monitor.slicer.zeta_centers[bid, n_slice] + dzeta / 2

        slice_mask = np.logical_and(zeta_coord < bin_max, zeta_coord >= bin_min)

        n_macroparticles_slice = np.sum(slice_mask)

        x_coord = (sigma_x * np.random.random(
            n_macroparticles_slice) + offs_x) * (bid + 1)
        px_coord = (sigma_px * np.random.random(
            n_macroparticles_slice) + offs_px) * (bid + 1)
        y_coord = (sigma_y * np.random.random(
            n_macroparticles_slice) + offs_y) * (bid + 1)
        py_coord = (sigma_py * np.random.random(
            n_macroparticles_slice) + offs_py) * (bid + 1)
        delta_coord = (sigma_delta * np.random.random(n_macroparticles_slice) +
                       offs_delta) * (bid + 1)

        x_coord_bunch.append(x_coord)
        px_coord_bunch.append(px_coord)
        y_coord_bunch.append(y_coord)
        py_coord_bunch.append(py_coord)
        delta_coord_bunch.append(delta_coord)
        zeta_coord_bunch.append(zeta_coord[slice_mask])
        n_macroparticles_slice_bunch.append(n_macroparticles_slice)

        particles.x[slice_mask] = x_coord
        particles.px[slice_mask] = px_coord
        particles.y[slice_mask] = y_coord
        particles.py[slice_mask] = py_coord
        particles.zeta[slice_mask] = zeta_coord[slice_mask]
        particles.delta[slice_mask] = delta_coord

    # track for twice flush_data_every turns so that we test the reshaping
    for _ in range(2 * flush_data_every):
        monitor.track(particles)

    with open(monitor._slices_filename, 'r') as f:
        bunches_data = json.load(f)

    for bid in bunch_numbers:
        slice_data = bunches_data[str(bid)]

        for turn in range(2 * flush_data_every):
            assert np.isclose(slice_data['mean_x'][turn][n_slice],
                              np.mean(x_coord_bunch[bid]))
            assert np.isclose(slice_data['mean_px'][turn][n_slice],
                              np.mean(px_coord_bunch[bid]))
            assert np.isclose(slice_data['mean_y'][turn][n_slice],
                              np.mean(y_coord_bunch[bid]))
            assert np.isclose(slice_data['mean_py'][turn][n_slice],
                              np.mean(py_coord_bunch[bid]))
            assert np.isclose(slice_data['mean_zeta'][turn][n_slice],
                              np.mean(zeta_coord_bunch[bid]))
            assert np.isclose(slice_data['mean_delta'][turn][n_slice],
                              np.mean(delta_coord_bunch[bid]))

            assert np.isclose(slice_data['sigma_x'][turn][n_slice],
                              np.std(x_coord_bunch[bid]))
            assert np.isclose(slice_data['sigma_px'][turn][n_slice],
                              np.std(px_coord_bunch[bid]))
            assert np.isclose(slice_data['sigma_y'][turn][n_slice],
                              np.std(y_coord_bunch[bid]))
            assert np.isclose(slice_data['sigma_py'][turn][n_slice],
                              np.std(py_coord_bunch[bid]))
            assert np.isclose(slice_data['sigma_zeta'][turn][n_slice],
                              np.std(zeta_coord_bunch[bid]))
            assert np.isclose(slice_data['sigma_delta'][turn][n_slice],
                              np.std(delta_coord_bunch[bid]))

            epsn_x = (np.sqrt(np.var(x_coord_bunch[bid]) *
                              np.var(px_coord_bunch[bid]) -
                              np.cov(x_coord_bunch[bid], px_coord_bunch[bid])[
                                  0, 1])
                      * beta * gamma)
            epsn_y = (np.sqrt(np.var(y_coord_bunch[bid]) *
                              np.var(py_coord_bunch[bid]) -
                              np.cov(y_coord_bunch[bid], py_coord_bunch[bid])[
                                  0, 1])
                      * beta * gamma)
            epsn_zeta = (np.sqrt(np.var(zeta_coord_bunch[bid]) *
                                 np.var(delta_coord_bunch[bid]) -
                                 np.cov(zeta_coord_bunch[bid],
                                        delta_coord_bunch[bid])[
                                     0, 1]) * beta * gamma)

            assert np.isclose(slice_data['epsn_x'][turn][n_slice], epsn_x)
            assert np.isclose(slice_data['epsn_y'][turn][n_slice], epsn_y)
            assert np.isclose(slice_data['epsn_zeta'][turn][n_slice], epsn_zeta)

            assert np.isclose(slice_data['num_particles'][turn][n_slice],
                              n_macroparticles_slice_bunch[bid])


@for_all_test_contexts
def test_particle_monitor_hdf5(test_context):
    n_macroparticles = int(1e2)
    zeta_range = (-1, 1)

    particles = xt.Particles(
        _context=test_context, p0c=7e12,
        x=np.random.random(n_macroparticles),
        px=np.random.random(n_macroparticles),
        y=np.random.random(n_macroparticles),
        py=np.random.random(n_macroparticles),
        zeta=np.random.random(n_macroparticles),
        delta=np.random.random(n_macroparticles),
    )

    # dummy particle mask
    particle_monitor_mask = particles.x > 0

    flush_data_every = 10

    monitor = xw.CollectiveMonitor(
        base_file_name='test_monitor',
        backend='hdf5',
        monitor_bunches=False,
        monitor_slices=False,
        monitor_particles=True,
        flush_data_every=flush_data_every,
        zeta_range=zeta_range,
        particle_monitor_mask=particle_monitor_mask,
    )

    # track for twice flush_data_every turns so that we test the reshaping
    for _ in range(2*flush_data_every):
        monitor.track(particles)

    with h5py.File(monitor._particles_filename, 'r') as h5file:
        for stat in monitor._stats_to_store_particles:
            saved_data = h5file[stat]
            for turn in range(2 * flush_data_every):
                assert np.allclose(getattr(particles,
                                           stat)[particle_monitor_mask],
                                   saved_data[turn, :])


@for_all_test_contexts
def test_particle_monitor_json(test_context):
    n_macroparticles = int(1e2)
    zeta_range = (-1, 1)

    particles = xt.Particles(
        _context=test_context, p0c=7e12,
        x=np.random.random(n_macroparticles),
        px=np.random.random(n_macroparticles),
        y=np.random.random(n_macroparticles),
        py=np.random.random(n_macroparticles),
        zeta=np.random.random(n_macroparticles),
        delta=np.random.random(n_macroparticles),
    )

    # dummy particle mask
    particle_monitor_mask = particles.x > 0

    flush_data_every = 10

    monitor = xw.CollectiveMonitor(
        base_file_name='test_monitor',
        backend='json',
        monitor_bunches=False,
        monitor_slices=False,
        monitor_particles=True,
        flush_data_every=flush_data_every,
        zeta_range=zeta_range,
        particle_monitor_mask=particle_monitor_mask,
    )

    # track for twice flush_data_every turns so that we test the reshaping
    for _ in range(2*flush_data_every):
        monitor.track(particles)

    with open(monitor._particles_filename, 'r') as f:
        buffer = json.load(f)

    for stat in monitor._stats_to_store_particles:
        saved_data = buffer[stat]
        for turn in range(2*flush_data_every):
            assert np.allclose(getattr(particles, stat)[particle_monitor_mask],
                               saved_data[turn][:])


@for_all_test_contexts
def test_flush_every(test_context):
    n_macroparticles = int(1e6)
    zeta_range = (-1, 1)

    energy = 7e12

    offs_x = 1e-3
    offs_px = 2e-3
    offs_y = 3e-3
    offs_py = 4e-3
    offs_zeta = 5e-3
    offs_delta = 6e-3

    sigma_x = 1e-3
    sigma_px = 2e-3
    sigma_y = 3e-3
    sigma_py = 4e-3
    sigma_zeta = 5e-3
    sigma_delta = 6e-3

    # base coordinates
    x_coord = sigma_x * np.random.random(n_macroparticles) + offs_x
    px_coord = sigma_px * np.random.random(n_macroparticles) + offs_px
    y_coord = sigma_y * np.random.random(n_macroparticles) + offs_y
    py_coord = sigma_py * np.random.random(n_macroparticles) + offs_py
    zeta_coord = sigma_zeta * np.random.random(n_macroparticles) + offs_zeta
    delta_coord = sigma_delta * np.random.random(n_macroparticles) + offs_delta

    # n_bunches bunches with n_macroparticles each with different coordinates
    particles = xt.Particles(
        _context=test_context, p0c=energy,
        x=x_coord,
        px=px_coord,
        y=y_coord,
        py=py_coord,
        zeta=zeta_coord,
        delta=delta_coord,
    )

    flush_data_every = 10

    monitor_10 = xw.CollectiveMonitor(
        base_file_name='test_monitor_10',
        backend='hdf5',
        monitor_bunches=True,
        monitor_slices=False,
        monitor_particles=False,
        flush_data_every=flush_data_every,
        zeta_range=zeta_range,
    )

    flush_data_every = 1

    monitor_1 = xw.CollectiveMonitor(
        base_file_name='test_monitor_1',
        backend='hdf5',
        monitor_bunches=True,
        monitor_slices=False,
        monitor_particles=False,
        flush_data_every=flush_data_every,
        zeta_range=zeta_range,
    )

    # track for twice flush_data_every turns so that we test the reshaping
    for _ in range(20):
        particles.x += 1
        monitor_10.track(particles)
        monitor_1.track(particles)

    dict_10 = {}
    dict_1 = {}

    with h5py.File(monitor_10._bunches_filename, 'r') as h5file:
        group_10 = h5file[str(0)]
        for kk in group_10.keys():
            dict_10[kk] = group_10[kk][:]

    with h5py.File(monitor_1._bunches_filename, 'r') as h5file:
        group_1 = h5file[str(0)]
        for kk in group_1.keys():
            dict_1[kk] = group_1[kk][:]

    for kk in dict_10.keys():
        assert np.allclose(dict_10[kk], dict_1[kk])



# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from typing import Tuple
import xpart as xp
import numpy as np

class BaseWake:
    pass

    def configure_for_tracking(self, zeta_range: Tuple[float, float],
                               num_slices: int,
                               num_turns=1,
                               bunch_spacing_zeta=None,  # This is P in the paper
                               filling_scheme=None,
                               circumference=None,
                               bunch_selection=None,
                               **kwargs # for multibunch compatibility
                               ) -> None:
        from xfields.beam_elements.waketracker import WakeTracker
        self._wake_tracker = WakeTracker(
            components=_expand_components(self.components),
            zeta_range=zeta_range,
            num_slices=num_slices,
            num_turns=num_turns,
            bunch_spacing_zeta=bunch_spacing_zeta,
            filling_scheme=filling_scheme,
            circumference=circumference,
            bunch_selection=bunch_selection,
            **kwargs)

    def _reconfigure_for_parallel(self, n_procs, my_rank) -> None:

        filled_slots = self._wake_tracker.slicer.filled_slots
        scheme = np.zeros(np.max(filled_slots) + 1,
                        dtype=np.int64)
        scheme[filled_slots] = 1

        split_scheme = xp.matched_gaussian.split_scheme
        bunch_selection_rank = split_scheme(filling_scheme=scheme,
                                             n_chunk=int(n_procs))

        self.configure_for_tracking(zeta_range=self._wake_tracker.zeta_range,
                          num_slices=self._wake_tracker.num_slices,
                          bunch_spacing_zeta=self._wake_tracker.bunch_spacing_zeta,
                          filling_scheme=scheme,
                          bunch_selection=bunch_selection_rank[my_rank],
                          num_turns=self._wake_tracker.num_turns,
                          circumference=self._wake_tracker.circumference,
                          )

    def track(self, particles) -> None:
        if not hasattr(self, '_wake_tracker') or self._wake_tracker is None:
            raise ValueError('Wake not configured for tracking, '
                             'call `configure_for_tracking` first')
        return self._wake_tracker.track(particles)


    def __add__(self, other):
        return _add_wakes(self, other)

    def __radd__(self, other):
        return _add_wakes(other, self)

    @property
    def slicer(self):
        if hasattr(self, '_wake_tracker') and self._wake_tracker is not None:
            return self._wake_tracker.slicer
        return None

def _add_wakes(wake1, wake2):
    out_components = []
    for ww in [wake1, wake2]:
        if ww is None or ww == 0:
            continue
        if isinstance(ww, CombinedWake):
            out_components.extend(ww.components)
        else:
            out_components.append(ww)

    return CombinedWake(components=out_components)


class CombinedWake(BaseWake):

    def __init__(self, components):
        self.components = components


def _handle_kind(kind):
    if isinstance(kind, str):
        kind = [kind]

    if isinstance(kind, (list, tuple)):
        kind = {kk: 1.0 for kk in kind}

    assert hasattr(kind, 'keys')

    return kind


def _expand_components(components):
    _expanded_components = []
    for cc in components:
        if hasattr(cc, 'function_vs_zeta'):
            _expanded_components.append(cc)
        else:
            assert hasattr(cc, 'components')
            _expanded_components.extend(
                _expand_components(cc.components))

    return _expanded_components



# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from __future__ import annotations

from .interface_dataclasses import Layer, FlatIW2DInput, RoundIW2DInput
from .parameters import *
from .utils import unique_sigfigs
from typing import Optional, Callable, Tuple, Union, List

import numpy as np
from scipy.constants import c as c_light, mu_0 as mu0, epsilon_0 as eps0
from numpy.typing import ArrayLike
from scipy import special as sp
from scipy.interpolate import interp1d

if hasattr(np, 'trapezoid'):
    trapz = np.trapezoid # numpy 2.0
else:
    trapz = np.trapz

Z0 = mu0 * c_light

KIND_DEFINITIONS = {
    'longitudinal':   {'plane': 'z', 'source_exponents':(0, 0), 'test_exponents': (0, 0)},
     'constant_x':    {'plane': 'x', 'source_exponents':(0, 0), 'test_exponents': (0, 0)},
     'constant_y':    {'plane': 'y', 'source_exponents':(0, 0), 'test_exponents': (0, 0)},
     'dipolar_x':      {'plane': 'x', 'source_exponents':(1, 0), 'test_exponents': (0, 0)},
     'dipolar_y':      {'plane': 'y', 'source_exponents':(0, 1), 'test_exponents': (0, 0)},
     'dipolar_xy':     {'plane': 'x', 'source_exponents':(0, 1), 'test_exponents': (0, 0)},
     'dipolar_yx':     {'plane': 'y', 'source_exponents':(1, 0), 'test_exponents': (0, 0)},
     'quadrupolar_x':  {'plane': 'x', 'source_exponents':(0, 0), 'test_exponents': (1, 0)},
     'quadrupolar_y':  {'plane': 'y', 'source_exponents':(0, 0), 'test_exponents': (0, 1)},
     'quadrupolar_xy': {'plane': 'x', 'source_exponents':(0, 0), 'test_exponents': (0, 1)},
     'quadrupolar_yx': {'plane': 'y', 'source_exponents':(0, 0), 'test_exponents': (1, 0)},
     'dipole_x':      {'plane': 'x', 'source_exponents':(1, 0), 'test_exponents': (0, 0)},
     'dipole_y':      {'plane': 'y', 'source_exponents':(0, 1), 'test_exponents': (0, 0)},
     'dipole_xy':     {'plane': 'x', 'source_exponents':(0, 1), 'test_exponents': (0, 0)},
     'dipole_yx':     {'plane': 'y', 'source_exponents':(1, 0), 'test_exponents': (0, 0)},
     'quadrupole_x':  {'plane': 'x', 'source_exponents':(0, 0), 'test_exponents': (1, 0)},
     'quadrupole_y':  {'plane': 'y', 'source_exponents':(0, 0), 'test_exponents': (0, 1)},
     'quadrupole_xy': {'plane': 'x', 'source_exponents':(0, 0), 'test_exponents': (0, 1)},
     'quadrupole_yx': {'plane': 'y', 'source_exponents':(0, 0), 'test_exponents': (1, 0)},
}


def mix_fine_and_rough_sampling(start: float, stop: float, rough_points: int,
                                fine_points: int, rois: List[Tuple[float, float]]):
    """
    Mix a fine and rough (geometric) sampling between start and stop,
    refined in the regions of interest rois.
    :param start: The lowest bound of the sampling
    :param stop: The highest bound of the sampling
    :param rough_points: The total number of data points to be
                         generated for the rough grid, between start
                         and stop.
    :param fine_points: The number of points in the fine grid of each roi.
    :param rois: List of unique tuples with the lower and upper bound of
                 of each region of interest.
    :return: An array with the sampling obtained.
    """
    intervals = [np.linspace(max(i, start), min(f, stop), fine_points)
                 for i, f in rois
                 if (start <= i <= stop or start <= f <= stop)]
    rough_sampling = np.geomspace(start, stop, rough_points)
    # the following concatenates adds the rough points to the fine sampling and sorts
    # the result. Then duplicates are removed, where two points are considered
    # equal if they are within 7 significant figures of each other.
    return unique_sigfigs(
        np.sort(np.hstack((*intervals, rough_sampling)),kind='stable'), 7)


class Component:
    """
    A data structure representing the impedance- and wake functions of some Component in a specified plane.
    """

    def __init__(self, impedance: Optional[Callable] = None, wake: Optional[Callable] = None,
                 plane: str = None,
                 source_exponents: Tuple[int, int] = (-1, -1), test_exponents: Tuple[int, int] = (-1, -1),
                 kind: str = None,
                 name: str = "Unnamed Component", f_rois: Optional[List[Tuple[float, float]]] = None,
                 t_rois: Optional[List[Tuple[float, float]]] = None):
        """
        The initialization function for the Component class.
        :param impedance: A callable function representing the impedance function of the Component. Can be undefined if
        the wake function is defined.
        :param wake: A callable function representing the wake function of the Component. Can be undefined if
        the impedance function is defined.
        :param plane: The plane of the Component, either 'x', 'y' or 'z'. Must be specified for valid initialization
        :param source_exponents: The exponents in the x and y planes experienced by the source particle. Also
        referred to as 'a' and 'b'. Must be specified for valid initialization
        :param test_exponents: The exponents in the x and y planes experienced by the source particle. Also
        referred to as 'a' and 'b'. Must be specified for valid initialization
        :param name: An optional user-specified name of the component
        :param f_rois: A list of tuples, each containing two floats, specifying the Regions Of Interest (ROIs) for the
        sampling of the impedance function. If not specified, upon discretization the impedance function will be
        sampled uniformly on a logarithmic scale.
        :param t_rois: A list of tuples, each containing two floats, specifying the Regions Of Interest (ROIs) for the
        sampling of the wake function. If not specified, upon discretization the wake function will be sampled
        uniformly on a logarithmic scale.
        """

        if kind is not None:
            assert kind in KIND_DEFINITIONS, (f"Invalid kind specified: {kind}."
                                            "Must be one of"
                                            f"{KIND_DEFINITIONS.keys()}")

        source_exponents, test_exponents, plane = _handle_plane_and_exponents_input(
                                    kind=kind, exponents=None,
                                    source_exponents=source_exponents,
                                    test_exponents=test_exponents,
                                    plane=plane)

        # Enforces that either impedance or wake is defined.
        assert impedance or wake, "The impedance and wake functions cannot both be undefined."
        # The impedance- and wake functions as callable objects, e.g lambda functions
        self._impedance = impedance
        self._wake = wake
        self.name = name

        # The plane of the Component, either 'x', 'y' or 'z'
        assert plane.lower() in ['x', 'y', 'z'], (
            f"Invalid plane specified: {plane}. Must be 'x', 'y' or 'z'.")

        self.plane = plane

        assert source_exponents != (-1, -1) and len(source_exponents) == 2, \
            "Cannot initialize Component object without specified source exponents (a, b)"
        self.source_exponents = source_exponents
        assert test_exponents != (-1, -1) and len(test_exponents) == 2, \
            "Cannot initialize Component object without specified test exponents (c, d)"
        self.test_exponents = test_exponents
        self.power_x = (source_exponents[0] + test_exponents[0] + (plane == 'x')) / 2
        self.power_y = (source_exponents[1] + test_exponents[1] + (plane == 'y')) / 2
        self._f_rois = f_rois if f_rois else []
        self._t_rois = t_rois if t_rois else []

    @property
    def impedance(self) -> Optional[Callable]:
        return self._impedance

    @property
    def wake(self) -> Optional[Callable]:
        return self._wake

    @property
    def f_rois(self) -> List[Tuple[float, float]]:
        return self._f_rois

    @property
    def t_rois(self) -> List[Tuple[float, float]]:
        return self._t_rois

    def function_vs_t(self, t, beta0, dt):
        out = self.wake(t)
        return out

    def function_vs_zeta(self, zeta, beta0, dzeta):
        out = self.function_vs_t(-zeta / beta0 / c_light, beta0,
                                 dzeta / beta0 / c_light)
        return out

    @property
    def kick(self) -> str:
        return {'x': 'px', 'y': 'py', 'z': 'delta'}[self.plane]

    @property
    def source_moments(self) -> List[str]:
        out = ['num_particles']
        if self.source_exponents[0] != 0:
            out.append('x')
        if self.source_exponents[1] != 0:
            out.append('y')

        return out

    def generate_wake_from_impedance(self) -> None:
        """
        Uses the impedance function of the Component object to generate its wake function, using
        a Fourier transform.
        :return: Nothing
        """
        # # If the object already has a wake function, there is no need to generate it.
        # if self.wake:
        #     pass
        # # In order to generate a wake function, we need to make sure the impedance function is defined.
        # assert self.impedance, "Tried to generate wake from impedance, but impedance is not defined."
        #
        # raise NotImplementedError

        # Temporary solution to avoid crashes
        self.wake = lambda x: 0

    def generate_impedance_from_wake(self) -> None:
        """
        Uses the wake function of the Component object to generate its impedance function, using
        a Fourier transform.
        :return: Nothing
        """
        # # If the object already has an impedance function, there is no need to generate it.
        # if self.impedance:
        #     pass
        # # In order to generate an impedance function, we need to make sure the wake function is defined.
        # assert self.wake, "Tried to generate impedance from wake, but wake is not defined."
        #
        # raise NotImplementedError

        # Temporary solution to avoid crashes
        self.impedance = lambda x: 0

    def is_compatible(self, other: Component) -> bool:
        """
        Compares all parameters of the self-object with the argument of the function and returns True if all of their
        attributes, apart from the impedance and wake functions, are identical.
        i.e. Returns True if self and other are compatible for Component addition, False otherwise
        :param other: Another Component
        :return: True if self and other can be added together, False otherwise
        """
        if not isinstance(other, Component):
            return False

        return all([self.source_exponents == other.source_exponents,
                    self.test_exponents == other.test_exponents,
                    self.plane == other.plane,
                    self.power_x == other.power_x,
                    self.power_y == other.power_y])

    def configure_for_tracking(self, zeta_range: Tuple[float, float],
                               num_slices: int,
                               **kwargs # for multibuunch compatibility
                               ) -> None:
        import xfields as xf
        self._xfields_wf = xf.Wakefield(components=[self], zeta_range=zeta_range,
                                        num_slices=num_slices, **kwargs)

    def track(self, particles):

        if not hasattr(self, '_xfields_wf') or not self._xfields_wf:
            raise ValueError('The component has not been configured for tracking. '
                             'Call configure_for_tracking before calling track.')
        self._xfields_wf.track(particles)


    def __add__(self, other: Component) -> Component:
        """
        Defines the addition operator for two Components
        :param self: The left addend
        :param other: The right addend
        :return: A new Component whose impedance and wake functions are the sums
        of the respective functions of the two addends.
        """

        if not isinstance(other, Component):
            return other.__radd__(self)

        # Enforce that the two addends are in the same plane
        assert self.plane == other.plane, "The two addends correspond to different planes and cannot be added.\n" \
                                          f"{self.plane} != {other.plane}"

        # Enforce that the two addends have the same exponent parameters
        assert self.source_exponents == other.source_exponents and self.test_exponents == other.test_exponents, \
            "The two addends have different exponent parameters and cannot be added.\n" \
            f"Source: {self.source_exponents} != {other.source_exponents}\n" \
            f"Test: {self.test_exponents} != {other.test_exponents}"

        # Defines an empty array to hold the two summed functions
        sums = []

        # Iterates through the two pairs of functions: impedances, then wakes
        for left, right in zip((self.impedance, self.wake), (other.impedance, other.wake)):
            # If neither addend has a defined function, we will not bother to calculate that of the sum
            if (not left) and (not right):
                sums.append(None)
            else:
                # # Generates the missing function for the addend which is missing it
                # if not left:
                #     [self.generate_impedance_from_wake, self.generate_wake_from_impedance][len(sums)]()
                # elif not right:
                #     [other.generate_impedance_from_wake, other.generate_wake_from_impedance][len(sums)]()
                #

                # TODO: Temporary fix until Fourier transform implemented
                if not left:
                    sums.append(right)
                elif not right:
                    sums.append(left)
                else:
                    # Appends the sum of the functions of the two addends to the list "sums"
                    sums.append(lambda x, l=left, r=right: l(x) + r(x))

        # Initializes and returns a new Component
        return Component(impedance=sums[0], wake=sums[1],
                         plane=self.plane,
                         source_exponents=self.source_exponents,
                         test_exponents=self.test_exponents,
                         f_rois=self.f_rois + other.f_rois,
                         t_rois=self.t_rois + other.t_rois)

    def __radd__(self, other: Union[int, Component]) -> Component:
        """
        Implements the __radd__ method for the Component class. This is only done to facilitate the syntactically
        practical use of the sum() method for Components. sum(iterable) works by adding all of the elements of the
        iterable to 0 sequentially. Thus, the behavior of the initial 0 + iterable[0] needs to be defined. In the case
        that the left addend of any addition involving a Component is 0, the resulting sum
        is simply defined to be the right addend.
        :param other: The left addend of an addition
        :return: The sum of self and other if other is a Component, otherwise just self.
        """

        # Checks if the left addend, other, is not a Component
        if other == 0:
            # In which case, the right addend is simply returned
            return self
        elif not isinstance(other, Component):
            return other.__add__(self)
        else:
        # Otherwise, their sum is returned (by invocation of Component.__add__(self, other))
            return self + other

    def __mul__(self, scalar: complex) -> Component:
        """
        Defines the behavior of multiplication of a Component by some, potentially complex, scalar
        :param scalar: A scalar value to be multiplied with some Component
        :return: A newly initialized Component identical to self in every way apart from the impedance-
        and wake functions, which have been multiplied by the scalar.
        """
        # Throws an AssertionError if scalar is not of the type complex, float or int
        assert isinstance(scalar, complex) or isinstance(scalar, float) or isinstance(scalar, int)
        # Initializes and returns a new Component with attributes like self, apart from the scaled functions
        return Component(impedance=(lambda x: scalar * self.impedance(x)) if self.impedance else None,
                         wake=(lambda x: scalar * self.wake(x)) if self.wake else None,
                         plane=self.plane,
                         source_exponents=self.source_exponents,
                         test_exponents=self.test_exponents, name=self.name,
                         f_rois=self.f_rois, t_rois=self.t_rois)

    def __rmul__(self, scalar: complex) -> Component:
        """
        Generalizes scalar multiplication of Component to be possibly from left and right. Both of these operations
        are identical.
        :param scalar: A scalar value to be multiplied with some Component
        :return: The result of calling Component.__mul__(self, scalar): A newly initialized Component identical to self
        in every way apart from the impedance- and wake functions, which have been multiplied by the scalar.
        """
        # Simply swaps the places of scalar and self in order to invoke the previously defined __mul__ function
        return self * scalar

    def __truediv__(self, scalar: complex) -> Component:
        """
        Implements the __truediv__ method for the Component class in order to produce the expected behavior when
        dividing a Component by some scalar. That is, the scalar multiplication of the multiplicative inverse of
        the scalar.
        :param scalar: A scalar value for the Component to be divided by
        :return: The result of calling Component.__mul__(self, 1 / scalar): A newly initialized Component identical to
        self in every way apart from the impedance- and wake functions, which have been multiplied by (1 / scalar).
        """
        # Defines the operation c / z to be equivalent to c * (1 / z) for some Component c and scalar z.
        return self * (1 / scalar)

    def __str__(self) -> str:
        """
        Implements the __str__ method for the Component class, providing an informative printout of the attributes
        of the Component.
        :return: A multi-line string containing information about the attributes of self, including which of the
        impedance- and wake functions are defined.
        """
        return f"{self.name} with parameters:\nPlane:\t\t\t{self.plane}\n" \
               f"Source exponents:\t{', '.join(str(i) for i in self.source_exponents)}\n" \
               f"Test exponents:\t\t{', '.join(str(i) for i in self.test_exponents)}\n" \
               f"Impedance function:\t{'DEFINED' if self.impedance else 'UNDEFINED'}\n" \
               f"Wake function:\t\t{'DEFINED' if self.wake else 'UNDEFINED'}\n" \
               f"Impedance-regions of interest: {', '.join(str(x) for x in self.f_rois)}\n" \
               f"Wake-regions of interest: {', '.join(str(x) for x in self.t_rois)}"

    def __lt__(self, other: Component) -> bool:
        """
        Implements the __lt__ (less than) method for the Component class. This has no real physical interpretation,
        it is just syntactically practical when iterating through the Components when adding two Elements.
        (see Element.__add__)
        :param other: The right hand side of a "<"-inequality
        :return: True if the self-Component is "less than" the other-Component by some arbitrary, but consistent,
        comparison.
        """
        # The two Components are compared by their attributes
        return [self.plane, self.source_exponents, self.test_exponents] < \
               [other.plane, other.source_exponents, other.test_exponents]

    def __eq__(self, other: Component) -> bool:
        """
        Implements the __eq__ method for the Component class. Two Components are designated as "equal" if the following
        two conditions hold:
        1. They are compatible for addition. That is, all of their attributes, apart from impedance- and wake functions,
        are exactly equal.
        2. The resulting values from evaluating the impedance functions of the two components respectively for 50
        points between 1 and 10000 all need to be close within some given tolerance. The same has to hold for the wake
        function.
        This somewhat approximated numerical approach to the equality comparator aims to compensate for small
        numerical/precision errors accumulated for two Components which have taken different "paths" to what should
        analytically be identical Components.
        :param other: The right hand side of the equality comparator
        :return: True if the two Components have identical attributes and sufficiently close functions, False otherwise
        """
        # First checks if the two Components are compatible for addition, i.e. if they have the same non-function
        # attributes
        if not self.is_compatible(other):
            # If they do not, they are not equal
            return False

        # Creates a numpy array of 50 points and verifies that the evaluations of the functions of the two components
        # for all of these components are sufficiently close. If they are, True is returned, otherwise, False is
        # returned.
        xs = np.linspace(1, 10000, 50)
        return (np.allclose(self.impedance(xs), other.impedance(xs), rtol=REL_TOL, atol=ABS_TOL) and
                np.allclose(self.wake(xs), other.wake(xs), rtol=REL_TOL, atol=ABS_TOL))

    def impedance_to_array(self, rough_points: int, start: float = MIN_FREQ,
                           stop: float = MAX_FREQ,
                           precision_factor: float = FREQ_P_FACTOR) -> Tuple[np.ndarray, np.ndarray]:
        """
        Produces a frequency grid based on the f_rois attribute of the component and evaluates the component's
        impedance function at these frequencies.
        :param rough_points: The total number of data points to be
                             generated for the rough grid, between start
                             and stop.
        :param start: The lowest frequency in the desired frequency grid
        :param stop: The highest frequency in the desired frequency grid
        :param precision_factor: A number indicating the ratio of points
               which should be placed within the regions of interest.
               If =0, the frequency grid will ignore the intervals in f_rois.
               If =1, the points will be distributed 1:1 between the
               rough grid and the fine grid on each roi. In general, =n
               means that there will be n times more points in the fine
               grid of each roi, than in the rough grid.
        :return: A tuple of two numpy arrays with same shape, giving
                 the frequency grid and impedances respectively
        """
        if len(self.f_rois) == 0:
            xs = np.geomspace(start, stop, rough_points)
            return xs, self.impedance(xs)

        # eliminate duplicates
        f_rois_no_dup = set(self.f_rois)

        fine_points_per_roi = int(round(rough_points * precision_factor))

        xs = mix_fine_and_rough_sampling(start, stop, rough_points,
                                         fine_points_per_roi,
                                         list(f_rois_no_dup))

        return xs, self.impedance(xs)

    def wake_to_array(self, rough_points: int, start: float = MIN_TIME,
                      stop: float = MAX_TIME,
                      precision_factor: float = TIME_P_FACTOR) -> Tuple[np.ndarray, np.ndarray]:
        """
        Produces a time grid based on the t_rois attribute of the component and evaluates the component's
        wake function at these time points.
        :param rough_points: The total number of data points to be
                             generated for the rough grid, between start
                             and stop.
        :param start: The lowest time in the desired time grid
        :param stop: The highest time in the desired time grid
        :param precision_factor: A number indicating the ratio of points
               which should be placed within the regions of interest.
               If =0, the frequency grid will ignore the intervals in t_rois.
               If =1, the points will be distributed 1:1 between the
               rough grid and the fine grid on each roi. In general, =n
               means that there will be n times more points in the fine
               grid of each roi, than in the rough grid.
        :return: A tuple of two numpy arrays with same shape, giving the
                 time grid and wakes respectively
        """
        if len(self.t_rois) == 0:
            xs = np.geomspace(start, stop, rough_points)
            return xs, self.wake(xs)

        # eliminate duplicates
        t_rois_no_dup = set(self.t_rois)

        fine_points_per_roi = int(round(rough_points * precision_factor))

        xs = mix_fine_and_rough_sampling(start, stop, rough_points,
                                         fine_points_per_roi,
                                         list(t_rois_no_dup))

        return xs, self.wake(xs)

    def discretize(self, freq_points: int, time_points: int, freq_start: float = MIN_FREQ, freq_stop: float = MAX_FREQ,
                   time_start: float = MIN_TIME, time_stop: float = MAX_TIME,
                   freq_precision_factor: float = FREQ_P_FACTOR,
                   time_precision_factor: float = TIME_P_FACTOR) -> Tuple[
        Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]:
        """
        Combines the two discretization-functions in order to fully discretize the wake and impedance of the object
        as specified by a number of parameters.
        :param freq_points: The total number of frequency/impedance points
        :param time_points: The total number of time/wake points
        :param freq_start: The lowest frequency in the frequency grid
        :param freq_stop: The highest frequency in the frequency grid
        :param time_start: The lowest time in the time grid
        :param time_stop: The highest time in the time grid
        :param freq_precision_factor: The ratio of points in the fine frequency grid
        (#points in frequency ROIs / rough frequency points)
        :param time_precision_factor: The ratio of points in the fine time grid
        (#points in time ROIs / rough time points)
        :return: A tuple of two tuples, containing the results of impedance_to_array and wake_to_array respectively
        """
        return (self.impedance_to_array(freq_points, freq_start, freq_stop, freq_precision_factor),
                self.wake_to_array(time_points, time_start, time_stop, time_precision_factor))

    def get_shorthand_type(self) -> str:
        """
        :return: A 5-character string indicating the plane as well as source- and test exponents of the component
        """
        return self.plane + "".join(str(x) for x in (self.source_exponents + self.test_exponents))


class ComponentResonator(Component):
    def __init__(self,
                kind: str = None,
                plane: str = None,
                exponents: Tuple[int, int, int, int]| None = None,
                source_exponents: Tuple[int, int] | None = None,
                test_exponents: Tuple[int, int] | None = None,
                r: float = None, q: float = None, f_r: float = None,
                f_roi_level: float = 0.5,
                factor=1.0) -> ComponentResonator:
        """
        Creates a resonator component with the given parameters
        :param plane: the plane the component corresponds to
        :param exponents: four integers corresponding to (source_x, source_y,
        test_x, test_y) aka (a, b, c, d)
        :param r: the shunt impedance of the given component of the resonator
        :param q: the quality factor of the given component of the resonator
        :param f_r: the resonance frequency of the given component of the
        resonator
        :param f_roi_level: fraction of the peak ok the resonator which is
        covered by the ROI. I.e. the roi will cover
        the frequencies for which the resonator impedance is larger than
        f_roi_level*r
        :return: A component object of a resonator, specified by the input
        arguments
        """
        self.r = r
        self.q = q
        self.f_r = f_r
        self.f_roi_level = f_roi_level
        self.factor = factor

        source_exponents, test_exponents, plane = _handle_plane_and_exponents_input(
                                    kind=kind, exponents=exponents,
                                    source_exponents=source_exponents,
                                    test_exponents=test_exponents,
                                    plane=plane)

        # we set impedance and wake to a dummy callable because they will be
        # overridden by methods
        super().__init__(impedance=lambda x: 0, wake=lambda x: 0, plane=plane,
                         source_exponents=source_exponents,
                         test_exponents=test_exponents,
                         name="Resonator")

    def impedance(self, f):
        f = np.atleast_1d(f)
        f_r = self.f_r
        q = self.q
        r = self.r
        factor = self.factor
        plane = self.plane
        if plane == 'z':
            out = factor * r / (1 - 1j * q * (f_r / f - f / f_r))
        else:
            out = factor * (f_r * r) / (f * (1 - 1j * q * (f_r / f - f / f_r)))

        return out


    def wake(self, t):
        t = np.atleast_1d(t)
        f_r = self.f_r
        q = self.q
        r = self.r
        factor = self.factor
        plane = self.plane
        omega_r = 2 * np.pi * f_r
        root_term = np.sqrt(1 - 1 / (4 * q ** 2) + 0J)
        omega_bar = omega_r * root_term
        alpha = omega_r / (2 * q)
        if plane == 'z':
            out = np.zeros_like(t)
            mask = t >= 0
            out[mask] = factor * (omega_r * r * np.exp(-alpha * t[mask]) * (
                   np.cos(omega_bar * t[mask]) -
                   alpha * np.sin(omega_bar * t[mask]) / omega_bar) / q).real
        else:
            out = np.zeros_like(t)
            mask = t >= 0
            out[mask] = factor * (omega_r * r * np.exp(-alpha * t[mask]) *
                   np.sin(omega_bar * t[mask]) /
                   (q * root_term)).real
        return out

    def function_vs_t(self, t, beta0, dt):
        out = self.wake(t)
        isscalar = np.isscalar(t)
        mask_left_edge = np.abs(t) < dt / 2
        # Handle discontinuity at t=0 consistently with beam loading theorem
        out[mask_left_edge] = 0.5 * self.wake(0)
        if isscalar:
            out = out[0]
        return out

    @property
    def f_rois(self):
        if self.q > 1:
            d = self._compute_resonator_f_roi_half_width(
                q=self.q, f_r=self.f_r, f_roi_level=self.f_roi_level)
            return [(self.f_r - d, self.f_r + d)]
        else:
            return []

    @property
    def t_rois(self):
        return []

    @staticmethod
    def _compute_resonator_f_roi_half_width(q: float, f_r: float,
                                            f_roi_level: float = 0.5):
        aux = np.sqrt((1 - f_roi_level) / f_roi_level)

        return (aux + np.sqrt(aux**2 + 4*q**2))*f_r/(2*q) - f_r


class ComponentClassicThickWall(Component):
    def __init__(self,
                 kind: str = None,
                 plane: str = None,
                 exponents: Tuple[int, int, int, int] | None = None,
                 source_exponents: Tuple[int, int] | None = None,
                 test_exponents: Tuple[int, int] | None = None,
                 layer: Layer = None,
                 radius: float = None,
                 resistivity: float = None,
                 factor: float = 1.0,
                 length: float = 1.0,
                 zero_rel_tol: float = 0.01
                 ) -> ComponentClassicThickWall:
        """
        Creates a single component object modeling a resistive wall
        impedance/wake, based on the "classic thick wall formula" (see e.g.
        A. W. Chao, chap. 2 in "Physics of Collective Beams Instabilities in
        High Energy Accelerators", John Wiley and Sons, 1993).
        Only longitudinal and transverse dipolar impedances are supported here.
        :param plane: the plane the component corresponds to
        :param exponents: four integers corresponding to (source_x, source_y,
        test_x, test_y) aka (a, b, c, d)
        :param layer: the chamber material, as a wit Layer object
        :param radius: the chamber radius in m
        :return: A component object
        """

        if (layer is None and resistivity is None):
            raise ValueError(
                "Either a layer object or a resistivity must be provided")
        if layer is not None and resistivity is not None:
            raise ValueError(
                "Both a layer object and a resistivity cannot be provided")

        if radius is None:
            raise ValueError("Radius must be provided")

        self.layer = layer
        self.radius = radius
        self.plane = plane
        self.factor = factor
        self.length = length
        self.resistivity = resistivity
        self.zero_rel_tol = zero_rel_tol

        source_exponents, test_exponents, plane = _handle_plane_and_exponents_input(
                                    kind=kind, exponents=exponents,
                                    source_exponents=source_exponents,
                                    test_exponents=test_exponents,
                                    plane=plane)

        # we set impedance and wake to a dummy callable because they will be
        # overridden by methods
        super().__init__(impedance=lambda x: 0, wake=lambda x: 0, plane=plane,
                         source_exponents=source_exponents,
                         test_exponents=test_exponents,
                         name="Classic Thick Wall")

    def impedance(self, f):
        layer = self.layer
        material_magnetic_permeability = mu0
        if layer is not None:
            self._check_layer(layer) # Checks that there are no unsupported layer properties
                                # (e.g. permeability different from vacuum)
            material_resistivity = layer.dc_resistivity
        else:
            material_resistivity = self.resistivity

        radius = self.radius
        plane = self.plane
        factor = self.factor
        length = self.length
        exponents = tuple(self.source_exponents + self.test_exponents)

        if plane == 'z' and exponents == (0, 0, 0, 0):
            out = factor * length * ((1 + np.sign(f)*1j) * material_resistivity
                            / (2* np.pi * radius)
                            / self.delta_skin(f, material_resistivity,
                                              material_magnetic_permeability))
        # Transverse dipolar impedance
        elif ((plane == 'x' and exponents == (1, 0, 0, 0)) or
                (plane == 'y' and exponents == (0, 1, 0, 0))):
            out = factor * c_light * length * ((1 + np.sign(f)*1j)
                * material_resistivity
                / (np.pi * radius**3 * (2 * np.pi * f ))
                / self.delta_skin(f, material_resistivity,
                                    material_magnetic_permeability))
        else:
            raise ValueError("Resistive wall wake not implemented for "
                  "component {}{}. Set to zero".format(plane, exponents))

        return out

    def wake(self, t):
        layer = self.layer
        if layer is not None:
            self._check_layer(layer) # Checks that there are no unsupported layer properties
                                # (e.g. permeability different from vacuum)
            material_resistivity = layer.dc_resistivity
        else:
            material_resistivity = self.resistivity
        radius = self.radius
        plane = self.plane
        factor = self.factor
        length = self.length
        exponents = tuple(self.source_exponents + self.test_exponents)

        isscalar = np.isscalar(t)
        t = np.atleast_1d(t)
        mask_positive = t > 1e-20
        out = np.zeros_like(t)

        # Longitudinal
        if plane == 'z' and exponents == (0, 0, 0, 0):
            out[mask_positive] = factor * length * (-1. / (4 * np.pi * radius)
                            * np.sqrt(Z0 * material_resistivity / np.pi / c_light)
                            * 1/(t[mask_positive]**(3/2)))
        # Transverse dipolar
        elif ((plane == 'x' and exponents == (1, 0, 0, 0)) or
                (plane == 'y' and exponents == (0, 1, 0, 0))):
            out[mask_positive] = factor * length * (
                        1. / (np.pi * radius**3)
                        * np.sqrt(c_light * Z0 * material_resistivity / np.pi)
                        * 1/(t[mask_positive]**(1/2)))
        else:
            raise ValueError("Resistive wall wake not implemented for "
                  "component {}{}. Set to zero".format(plane, exponents))

        if isscalar:
            out = out[0]

        return out

    @staticmethod
    def delta_skin(f, material_resistivity, material_magnetic_permeability):
        return np.sqrt(2 * material_resistivity / (2*np.pi*abs(f) *
                                        material_magnetic_permeability))

    @property
    def f_rois(self):
        return []

    @property
    def t_rois(self):
        return []

    def function_vs_t(self, t, beta0, dt):

        isscalar = np.isscalar(t)
        t = np.atleast_1d(t)

        assert dt > 0
        mask_zero = np.abs(t) < dt * self.zero_rel_tol
        out = np.zeros_like(t)
        out[mask_zero] = self.wake(dt * self.zero_rel_tol)
        out[~mask_zero] = self.wake(t[~mask_zero])

        if isscalar:
            out = out[0]
        return out

    @staticmethod
    def _check_layer(layer):
        if not isinstance(layer, Layer):
            raise ValueError("Layer must be a wit Layer object")
        if layer.thickness is not None:
            raise ValueError("Layer thickness not supported for "
                             "classic thick wall impedance")
        if layer.re_dielectric_constant != 1.0:
            raise ValueError("Dielectric constant not supported for "
                             "classic thick wall impedance")
        if layer.permeability_relaxation_frequency < np.inf:
            raise ValueError("Permeability relaxation frequency not supported for "
                             "classic thick wall impedance")


class ComponentSingleLayerResistiveWall(Component):
    def __init__(self,
                 kind: str = None,
                 plane: str = None,
                 exponents: Tuple[int, int, int, int] = None,
                 source_exponents: Tuple[int, int] | None = None,
                 test_exponents: Tuple[int, int] | None = None,
                 input_data: Union[FlatIW2DInput, RoundIW2DInput] = None,
                 factor: float = 1.0):
        """
        Creates a single component object modeling a resistive wall impedance,
        based on the single-layer approximated formulas by E. Metral (see e.g.
        Eqs. 13-14 in N. Mounet and E. Metral, IPAC'10, TUPD053,
        https://accelconf.web.cern.ch/IPAC10/papers/tupd053.pdf, and
        Eq. 21 in F. Roncarolo et al, Phys. Rev. ST Accel. Beams 12, 084401,
        2009, https://doi.org/10.1103/PhysRevSTAB.12.084401)
        :param plane: the plane the component corresponds to
        :param exponents: four integers corresponding to (source_x, source_y,
        test_x, test_y) aka (a, b, c, d)
        :param input_data: an IW2D input object (flat or round). If the input
        is of type FlatIW2DInput and symmetric, we apply to the round formula
        the Yokoya factors for an infinitely flat structure (see e.g. K. Yokoya,
        KEK Preprint 92-196 (1993), and Part. Accel. 41 (1993) pp.221-248,
        https://cds.cern.ch/record/248630/files/p221.pdf),
        while for a single plate we use those from A. Burov and V. Danilov,
        PRL 82,11 (1999), https://doi.org/10.1103/PhysRevLett.82.2286. Other
        kinds of asymmetric structure will raise an error.
        If the input is of type RoundIW2DInput, the structure is in principle
        round but the Yokoya factors put in the input will be used.
        :return: A component object
        """
        self.input_data = input_data
        self.plane = plane
        self.factor = factor

        source_exponents, test_exponents, plane = _handle_plane_and_exponents_input(
                                    kind=kind, exponents=exponents,
                                    source_exponents=source_exponents,
                                    test_exponents=test_exponents,
                                    plane=plane)
        factor = self.factor

        if isinstance(input_data, FlatIW2DInput):
            if len(input_data.top_layers) > 1:
                raise NotImplementedError("Input data can have only one layer")
            self.yok_long = 1.
            self.layer = input_data.top_layers[0]
            self.radius = input_data.top_half_gap
            if input_data.top_bottom_symmetry:
                self.yok_dipx = np.pi**2/24.
                self.yok_dipy = np.pi**2/12.
                self.yok_quax = -np.pi**2/24.
                self.yok_quay = np.pi**2/24.
            elif input_data.bottom_half_gap == np.inf:
                self.yok_dipx = 0.25
                self.yok_dipy = 0.25
                self.yok_quax = -0.25
                self.yok_quay = 0.25
            else:
                raise NotImplementedError("For asymmetric structures, only the "
                                          "case of a single plate is "
                                          "implemented; hence the bottom "
                                          "half-gap must be infinite")
        elif isinstance(input_data, RoundIW2DInput):
            self.radius = input_data.inner_layer_radius
            if len(input_data.layers) > 1:
                raise NotImplementedError("Input data can have only one layer")
            self.layer = input_data.layers[0]
            self.yok_long = input_data.yokoya_factors[0]
            self.yok_dipx = input_data.yokoya_factors[1]
            self.yok_dipy = input_data.yokoya_factors[2]
            self.yok_quax = input_data.yokoya_factors[3]
            self.yok_quay = input_data.yokoya_factors[4]
        else:
            raise NotImplementedError("Input of type neither FlatIW2DInput nor "
                                       "RoundIW2DInput cannot be handled")


        # we set impedance and wake to a dummy callable because they will be
        # overridden by methods
        super().__init__(impedance=lambda x: 0, wake=lambda x: 0, plane=plane,
                         source_exponents=source_exponents,
                         test_exponents=test_exponents,
                         name="Single layer resistive wall")

    def impedance(self, f):

        factor = self.factor

        # Longitudinal impedance
        if (self.plane == 'z' and self.source_exponents == (0, 0)
           and self.test_exponents == (0, 0)):
            out = factor * self.yok_long*self._zlong_round_single_layer_approx(
                                    frequencies=f,
                                    gamma=self.input_data.relativistic_gamma,
                                    layer=self.layer,
                                    radius=self.radius,
                                    length=self.input_data.length)
        # Transverse impedances
        elif (self.plane == 'x' and self.source_exponents == (1, 0) and
              self.test_exponents == (0, 0)):
            out = factor * self.yok_dipx * self._zdip_round_single_layer_approx(
                frequencies=f,
                gamma=self.input_data.relativistic_gamma,
                layer=self.layer,
                radius=self.radius,
                length=self.input_data.length)
        elif (self.plane == 'y' and self.source_exponents == (0, 1) and
              self.test_exponents == (0, 0)):
            out = factor * self.yok_dipy * self._zdip_round_single_layer_approx(
                frequencies=f,
                gamma=self.input_data.relativistic_gamma,
                layer=self.layer,
                radius=self.radius,
                length=self.input_data.length)
        elif (self.plane == 'x' and self.source_exponents == (0, 0)
             and self.test_exponents == (1, 0)):
            out = factor * self.yok_quax * self._zdip_round_single_layer_approx(
                frequencies=f,
                gamma=self.input_data.relativistic_gamma,
                layer=self.layer,
                radius=self.radius,
                length=self.input_data.length)
        elif (self.plane == 'y' and self.source_exponents == (0, 0) and
              self.test_exponents == (0, 1)):
            out = factor * self.yok_quay * self._zdip_round_single_layer_approx(
                frequencies=f,
                gamma=self.input_data.relativistic_gamma,
                layer=self.layer,
                radius=self.radius,
                length=self.input_data.length)
        else:
            out = np.zeros_like(f)

        return out

    def wake(self, t):
        raise NotImplementedError("Wake not implemented for single-layer "
                                  "resistive wall impedance")

    @staticmethod
    def _zlong_round_single_layer_approx(frequencies: ArrayLike, gamma: float,
                                         layer: Layer, radius: float,
                                         length: float) -> ArrayLike:
        """
        Function to compute the longitudinal resistive-wall impedance from
        the single-layer, approximated formula for a cylindrical structure,
        by E. Metral (see e.g. Eqs. 13-14 in N. Mounet and E. Metral, IPAC'10,
        TUPD053, https://accelconf.web.cern.ch/IPAC10/papers/tupd053.pdf, and
        Eq. 21 in F. Roncarolo et al, Phys. Rev. ST Accel. Beams 12, 084401,
        2009, https://doi.org/10.1103/PhysRevSTAB.12.084401)
        :param frequencies: the frequencies (array) (in Hz)
        :param gamma: relativistic mass factor
        :param layer: a layer with material properties (only resistivity,
        relaxation time and magnetic susceptibility are taken into account
        at this stage)
        :param radius: the radius of the structure (in m)
        :param length: the total length of the resistive object (in m)
        :return: Zlong, the longitudinal impedance at these frequencies
        """
        beta = np.sqrt(1 - 1 / gamma**2)
        omega = 2 * np.pi * frequencies
        k = omega / (beta * c_light)

        rho = layer.dc_resistivity
        tau = layer.resistivity_relaxation_time
        mu1 = 1 + layer.magnetic_susceptibility
        eps1 = 1 - 1j / (eps0 * rho * omega * (1 + 1j * omega * tau))
        nu = k * np.sqrt(1 - beta**2 * eps1 * mu1)

        coef_long = 1j * omega * mu0 * length / (2 * np.pi * beta**2 * gamma**2)

        x1 = k * radius / gamma
        x1sq = x1**2
        x2 = nu * radius

        zlong = coef_long * (sp.k0(x1) / sp.i0(x1) -
                             1 / (x1sq * (1/2 + eps1 * sp.kve(1, x2)/
                                          (x2 * sp.kve(0, x2)))))

        return zlong

    @staticmethod
    def _zdip_round_single_layer_approx(frequencies: ArrayLike, gamma: float,
                                        layer: Layer, radius: float,
                                        length: float) -> ArrayLike:
        """
        Function to compute the transverse dipolar resistive-wall impedance from
        the single-layer, approximated formula for a cylindrical structure,
        Eqs. 13-14 in N. Mounet and E. Metral, IPAC'10, TUPD053,
        https://accelconf.web.cern.ch/IPAC10/papers/tupd053.pdf, and
        Eq. 21 in F. Roncarolo et al, Phys. Rev. ST Accel. Beams 12, 084401,
        2009, https://doi.org/10.1103/PhysRevSTAB.12.084401)
        :param frequencies: the frequencies (array) (in Hz)
        :param gamma: relativistic mass factor
        :param layer: a layer with material properties (only resistivity,
        relaxation time and magnetic susceptibility are taken into account
        at this stage)
        :param radius: the radius of the structure (in m)
        :param length: the total length of the resistive object (in m)
        :return: Zdip, the transverse dipolar impedance at these frequencies
        """
        beta = np.sqrt(1 - 1 / gamma**2)
        omega = 2 * np.pi * frequencies
        k = omega / (beta * c_light)

        rho = layer.dc_resistivity
        tau = layer.resistivity_relaxation_time
        mu1 = 1 + layer.magnetic_susceptibility
        eps1 = 1 - 1j/(eps0 * rho * omega*(1 + 1j * omega * tau))
        nu = k * np.sqrt(1 - beta**2 * eps1 * mu1)

        coef_dip = 1j * k**2 * Z0 * length/(4. * np.pi * beta * gamma**4)

        x1 = k * radius / gamma
        x1sq = x1**2
        x2 = nu * radius

        zdip = coef_dip * (sp.k1(x1) / sp.i1(x1) +
                           4 * beta**2 * gamma**2 /
                           (x1sq*(2 + x2 * sp.kve(0, x2) /
                                  (mu1 * sp.kve(1, x2)))))

        return zdip


class ComponentTaperSingleLayerRestsistiveWall(Component):
    def __init__(self,
                 kind: str = None,
                 plane: str = None,
                 exponents: Tuple[int, int, int, int] = None,
                 source_exponents: Tuple[int, int] | None = None,
                 test_exponents: Tuple[int, int] | None = None,
                 input_data: Union[FlatIW2DInput, RoundIW2DInput] = None,
                 radius_small: float = None, radius_large: float = None,
                 step_size: float = 1e-3,
                 factor: float = 1.0) -> ComponentTaperSingleLayerRestsistiveWall:
        """
        Creates a single component object modeling a round or flat taper (flatness
        along the horizontal direction, change of half-gap along the vertical one)
        resistive-wall impedance, using the integration of the radius-dependent
        approximated formula for a cylindrical structure (see
        the above functions), over the length of the taper.
        :param plane: the plane the component corresponds to
        :param exponents: four integers corresponding to (source_x, source_y, test_x, test_y) aka (a, b, c, d)
        :param input_data: an IW2D input object (flat or round). If the input
        is of type FlatIW2DInput and symmetric, we apply to the round formula the
        Yokoya factors for an infinitely flat structure (see e.g. K. Yokoya,
        KEK Preprint 92-196 (1993), and Part. Accel. 41 (1993) pp.221-248,
        https://cds.cern.ch/record/248630/files/p221.pdf),
        while for a single plate we use those from A. Burov and V. Danilov,
        PRL 82,11 (1999), https://doi.org/10.1103/PhysRevLett.82.2286. Other
        kinds of asymmetric structure will raise an error.
        If the input is of type RoundIW2DInput, the structure is in principle
        round but the Yokoya factors put in the input will be used.
        Note that the radius or half-gaps in input_data are not used (replaced
        by the scan from radius_small to radius_large, for the integration).
        :param radius_small: the smallest radius of the taper (in m)
        :param radius_large: the largest radius of the taper (in m)
        :param step_size: the step size (in the radial or vertical direction)
        for the integration (in m)
        :return: A component object
        """
        self.input_data = input_data
        self.radius_large = radius_large
        self.radius_small = radius_small
        self.step_size = step_size
        self.plane = plane
        self.factor = factor

        source_exponents, test_exponents, plane = _handle_plane_and_exponents_input(
                                    kind=kind, exponents=exponents,
                                    source_exponents=source_exponents,
                                    test_exponents=test_exponents,
                                    plane=plane)

        if isinstance(input_data, FlatIW2DInput):
            if len(input_data.top_layers) > 1:
                raise NotImplementedError("Input data can have only one layer")
            self.yok_long = 1.
            self.layer = input_data.top_layers[0]
            self.radius = input_data.top_half_gap
            if input_data.top_bottom_symmetry:
                self.yok_dipx = np.pi**2/24.
                self.yok_dipy = np.pi**2/12.
                self.yok_quax = -np.pi**2/24.
                self.yok_quay = np.pi**2/24.
            elif input_data.bottom_half_gap == np.inf:
                self.yok_dipx = 0.25
                self.yok_dipy = 0.25
                self.yok_quax = -0.25
                self.yok_quay = 0.25
            else:
                raise NotImplementedError("For asymmetric structures, only the case of a single plate is implemented; "
                                        "hence the bottom half gap must be infinite")
        elif isinstance(input_data, RoundIW2DInput):
            self.radius = input_data.inner_layer_radius
            if len(input_data.layers) > 1:
                raise NotImplementedError("Input data can have only one layer")
            self.layer = input_data.layers[0]
            self.yok_long = input_data.yokoya_factors[0]
            self.yok_dipx = input_data.yokoya_factors[1]
            self.yok_dipy = input_data.yokoya_factors[2]
            self.yok_quax = input_data.yokoya_factors[3]
            self.yok_quay = input_data.yokoya_factors[4]
        else:
            raise NotImplementedError("Input of type neither FlatIW2DInput nor RoundIW2DInput cannot be handled")

        # we set impedance and wake to a dummy callable because they will be
        # overridden by methods
        super().__init__(impedance=lambda x: 0, wake=lambda x: 0, plane=plane,
                         source_exponents=source_exponents,
                         test_exponents=test_exponents,
                         name="Taper single layer resistive wall")

    def impedance(self, f):

        factor = self.factor
        # Longitudinal impedance
        if (self.plane == 'z' and self.source_exponents == (0, 0)
            and self.test_exponents == (0, 0)):
            out = factor * self.yok_long*self._zlong_round_taper_RW_approx(
                            frequencies=f,
                            gamma=self.input_data.relativistic_gamma,
                            layer=self.layer,
                            radius_small=self.radius_small,
                            radius_large=self.radius_large,
                            length=self.input_data.length,
                            step_size=self.step_size)
        # Transverse impedances
        elif (self.plane == 'x' and self.source_exponents == (1, 0)
              and self.test_exponents == (0, 0)):
            out = factor * self.yok_dipx*self._zdip_round_taper_RW_approx(
                            frequencies=f,
                            gamma=self.input_data.relativistic_gamma,
                            layer=self.layer,
                            radius_small=self.radius_small,
                            radius_large=self.radius_large,
                            length=self.input_data.length,
                            step_size=self.step_size)
        elif (self.plane == 'y' and self.source_exponents == (0, 1)
              and self.test_exponents == (0, 0)):
            out = factor * self.yok_dipy*self._zdip_round_taper_RW_approx(
                            frequencies=f,
                            gamma=self.input_data.relativistic_gamma,
                            layer=self.layer,
                            radius_small=self.radius_small,
                            radius_large=self.radius_large,
                            length=self.input_data.length,
                            step_size=self.step_size)
        elif (self.plane == 'x' and self.source_exponents == (0, 0)
             and self.test_exponents == (1, 0)):
            out = factor * self.yok_quax*self._zdip_round_taper_RW_approx(
                            frequencies=f,
                            gamma=self.input_data.relativistic_gamma,
                            layer=self.layer,
                            radius_small=self.radius_small,
                            radius_large=self.radius_large,
                            length=self.input_data.length,
                            step_size=self.step_size)
        elif (self.plane == 'y' and self.source_exponents == (0, 0)
              and self.test_exponents == (0, 1)):
            out = factor * self.yok_quay*self._zdip_round_taper_RW_approx(
                            frequencies=f,
                            gamma=self.input_data.relativistic_gamma,
                            layer=self.layer,
                            radius_small=self.radius_small,
                            radius_large=self.radius_large,
                            length=self.input_data.length,
                            step_size=self.step_size)
        else:
            out = np.zeros_like(f)

        return out


    def wake(self, t):
        raise NotImplementedError("Wake not implemented for single-layer "
                                  "resistive wall impedance")


    @staticmethod
    def _zlong_round_taper_RW_approx(frequencies: ArrayLike, gamma: float,
                                    layer: Layer, radius_small: float,
                                    radius_large: float, length: float,
                                    step_size: float = 1e-3) -> ArrayLike:
        """
        Function to compute the longitudinal resistive-wall impedance for a
        round taper, integrating the radius-dependent approximated formula
        for a cylindrical structure (see_zlong_round_single_layer_approx above),
        over the length of the taper.
        :param frequencies: the frequencies (array) (in Hz)
        :param gamma: relativistic mass factor
        :param layer: a layer with material properties (only resistivity,
        relaxation time and magnetic susceptibility are taken into account
        at this stage)
        :param radius_small: the smallest radius of the taper (in m)
        :param radius_large: the largest radius of the taper (in m)
        :param length: the total length of the taper (in m)
        :param step_size: the step size (in the radial direction) for the
        integration (in m)
        :return: the longitudinal impedance at these frequencies
        """
        if np.isscalar(frequencies):
            frequencies = np.array(frequencies)
        beta = np.sqrt(1.-1./gamma**2)
        omega = 2*np.pi*frequencies.reshape((-1, 1))
        k = omega/(beta*c_light)

        rho = layer.dc_resistivity
        tau = layer.resistivity_relaxation_time
        mu1 = 1.+layer.magnetic_susceptibility
        eps1 = 1. - 1j/(eps0*rho*omega*(1.+1j*omega*tau))
        nu = k*np.sqrt(1.-beta**2*eps1*mu1)

        coef_long = 1j*omega*mu0/(2.*np.pi*beta**2*gamma**2)

        npts = int(np.floor(abs(radius_large-radius_small)/step_size)+1)
        radii = np.linspace(radius_small, radius_large, npts).reshape((1, -1))
        one_array = np.ones(radii.shape)

        x1 = k.dot(radii)/gamma
        x1sq = x1**2
        x2 = nu.dot(radii)
        zlong = (coef_long.dot(length / float(npts) * one_array) *
                (sp.k0(x1) / sp.i0(x1) -
                 1. / (x1sq * (1. / 2. + eps1.dot(one_array) *
                               sp.kve(1, x2) / (x2 * sp.kve(0, x2)))))
                )

        return trapz(zlong, axis=1)


    @staticmethod
    def _zdip_round_taper_RW_approx(frequencies: ArrayLike, gamma: float,
                                    layer: Layer, radius_small: float,
                                    radius_large: float, length: float,
                                    step_size: float = 1e-3) -> ArrayLike:
        """
        Function to compute the transverse dip. resistive-wall impedance for a
        round taper, integrating the radius-dependent approximated formula
        for a cylindrical structure (see_zdip_round_single_layer_approx above),
        over the length of the taper.
        :param frequencies: the frequencies (array) (in Hz)
        :param gamma: relativistic mass factor
        :param layer: a layer with material properties (only resistivity,
        relaxation time and magnetic susceptibility are taken into account
        at this stage)
        :param radius_small: the smallest radius of the taper (in m)
        :param radius_large: the largest radius of the taper (in m)
        :param length: the total length of the taper (in m)
        :param step_size: the step size (in the radial direction) for the
        integration (in m)
        :return: the transverse dipolar impedance at these frequencies
        """
        if np.isscalar(frequencies):
            frequencies = np.array(frequencies)
        beta = np.sqrt(1.-1./gamma**2)
        omega = 2*np.pi*frequencies.reshape((-1,1))
        k = omega/(beta*c_light)

        rho = layer.dc_resistivity
        tau = layer.resistivity_relaxation_time
        mu1 = 1.+layer.magnetic_susceptibility
        eps1 = 1. - 1j/(eps0*rho*omega*(1.+1j*omega*tau))
        nu = k*np.sqrt(1.-beta**2*eps1*mu1)

        coef_dip = 1j*k**2*Z0/(4.*np.pi*beta*gamma**4)

        npts = int(np.floor(abs(radius_large-radius_small)/step_size)+1)
        radii = np.linspace(radius_small,radius_large,npts).reshape((1,-1))
        one_array = np.ones(radii.shape)

        x1 = k.dot(radii)/gamma
        x1sq = x1**2
        x2 = nu.dot(radii)
        zdip = (
                coef_dip.dot(length / float(npts) * one_array) *
                (sp.k1(x1) / sp.i1(x1) +
                 4 * beta**2 * gamma**2 / (x1sq * (2 + x2 * sp.kve(0, x2) /
                                                   (mu1 * sp.kve(1, x2)))))
            )

        return trapz(zdip, axis=1)


class ComponentInterpolated(Component):
    """
    Creates a component in which the impedance function is evaluated directly
    only on few points and it is interpolated everywhere else. This helps when
    the impedance function is very slow to evaluate.
    """
    def __init__(self,
                interpolation_frequencies: ArrayLike = None,
                impedance_input: Optional[Callable] = None,
                interpolation_times: ArrayLike = None,
                wake_input: Optional[Callable] = None,
                kind: str = None,
                plane: str = None,
                source_exponents: Tuple[int, int] = None,
                test_exponents: Tuple[int, int] = None,
                name: str = "Interpolated Component",
                f_rois: Optional[List[Tuple[float, float]]] = None,
                t_rois: Optional[List[Tuple[float, float]]] = None):
        """
        The init of the ComponentInterpolated class

        :param interpolation_frequencies: the frequencies where the impedance
        function is evaluated for the interpolation
        :param impedance_input: A callable function representing the impedance
        function of the Component. Can be undefined if the wake function is defined
        :param interpolation_times: the times where the wake function is evaluated
        for the interpolation
        :param wake_input: A callable function representing the wake function of the
        Component. Can be undefined if the impedance function is defined.
        :param plane: The plane of the Component, either 'x', 'y' or 'z'. Must be
        specified for valid initialization
        :param source_exponents: The exponents in the x and y planes experienced by
        the source particle
        :param test_exponents: The exponents in the x and y planes experienced by
        the test particle
        :param name: An optional user-specified name of the component
        :param f_rois: a list of frequency regions of interest
        :param t_rois: a list of time regions of interest
        """
        assert ((interpolation_frequencies is not None) ==
                (impedance_input is not None)), ("Either both or none of the "
                "impedance and the interpolation frequencies must be given")

        source_exponents, test_exponents, plane = _handle_plane_and_exponents_input(
                            kind=kind, exponents=None,
                            source_exponents=source_exponents,
                            test_exponents=test_exponents,
                            plane=plane)

        self.interpolation_frequencies = interpolation_frequencies
        self.impedance_samples = impedance_input(self.interpolation_frequencies)

        assert ((interpolation_times is not None) ==
                (wake_input is not None)), ("Either both or none of the wake "
                "and the interpolation times must be given")

        self.interpolation_times = interpolation_times
        self.wake_samples = wake_input(self.interpolation_times)

        super().__init__(impedance=lambda x: 0, wake=lambda x: 0, plane=plane,
                       source_exponents=source_exponents,
                       test_exponents=test_exponents,
                       f_rois=f_rois, t_rois=t_rois,
                       name=name)

    def impedance(self, f):
        if hasattr(self, 'impedance_samples') is not None:
            return np.interp(f, self.interpolation_frequencies,
                             self.impedance_samples)
        else:
            return np.zeros_like(f)

    def wake(self, t):
        if hasattr(self, 'wake_samples'):
            return np.interp(t, self.interpolation_times, self.wake_samples,
                             left=0, right=0 # pad with zeros outside the range
            )
        else:
            return np.zeros_like(t)


class ComponentFromArrays(Component):
    """
    A component from impedance and/or wake functions defined discretely
    through arrays
    """
    def __init__(self,
                 interpolation_frequencies: ArrayLike = None,
                 impedance_samples: ArrayLike = None,
                 interpolation_times: ArrayLike = None,
                 wake_samples: ArrayLike = None,
                 kind: str = None,
                 plane: str = None,
                 source_exponents: Tuple[int, int] = None,
                 test_exponents: Tuple[int, int] = None,
                 name: str = "Interpolated Component",
                 f_rois: Optional[List[Tuple[float, float]]] = None,
                 t_rois: Optional[List[Tuple[float, float]]] = None):
        """
        The init of the ComponentFromArrays class

        :param interpolation_frequencies: the frequencies where the impedance
        function is evaluated for the interpolation
        :param impedance_samples: Aan array of impedance values at the interpolation
        frequencies
        the wake function is defined.
        :param interpolation_times: the times where the wake function is evaluated
        for the interpolation
        :param wake_samples: an array of wake values at the interpolation times
        Component. Can be undefined if the impedance function is defined.
        :param plane: The plane of the Component, either 'x', 'y' or 'z'. Must be
        specified for valid initialization
        :param source_exponents: The exponents in the x and y planes experienced by
        the source particle
        :param test_exponents: The exponents in the x and y planes experienced by
        the test particle
        :param name: An optional user-specified name of the component
        :param f_rois: a list of frequency regions of interest
        :param t_rois: a list of time regions of interest
        """
        source_exponents, test_exponents, plane = _handle_plane_and_exponents_input(
                            kind=kind, exponents=None,
                            source_exponents=source_exponents,
                            test_exponents=test_exponents,
                            plane=plane)

        self.interpolation_frequencies = interpolation_frequencies
        self.impedance_samples = impedance_samples

        self.interpolation_times = interpolation_times
        self.wake_samples = wake_samples

        if self.interpolation_frequencies is not None:
            impedance = interp1d(self.interpolation_frequencies,
                                 self.impedance_samples)
        else:
            impedance = lambda f: 0

        if self.interpolation_times is not None:
            wake = interp1d(self.interpolation_times,
                            self.wake_samples,
                            # pad with zeros outside the range
                            fill_value=0, bounds_error=False
                            )
        else:
            wake = lambda t: 0

        super().__init__(impedance=impedance, wake=wake, plane=plane,
                         source_exponents=source_exponents,
                         test_exponents=test_exponents,
                         f_rois=f_rois, t_rois=t_rois,
                         name=name)

    def function_vs_t(self, t, beta0, dt):

        assert dt > 0

        isscalar = np.isscalar(t)
        if isscalar:
            t = np.array([t])

        out = self.wake(t)

        # At the edges of the provided wake table we take half the provided
        # value. This is equivalent to assume the next sample (not provided) is
        # zero. The advantage is that for longitudinal ultra-relativistic wakes
        # that have a discontinuity in zero, it provides a kick consistent with
        # the fundamental theorem of beam loading (see A. Chao, Physics of
        # Collective Beam Instabilities in High Energy Accelerators, Wiley, 
        # 1993, Fig.2.6)
        mask_left_edge = np.abs(t - self.interpolation_times[0]) < dt / 2
        out[mask_left_edge] = self.wake_samples[0] / 2.
        mask_right_edge = np.abs(t - self.interpolation_times[-1]) < dt / 2
        out[mask_right_edge] = self.wake_samples[-1] / 2.

        if isscalar:
            out = out[0]

        return out



def _handle_plane_and_exponents_input(kind, exponents, source_exponents, test_exponents, plane):

    if kind is not None:
        assert exponents is None, (
            "If `kind` is specified, `exponents` should not be specified.")
        assert source_exponents is None and test_exponents is None, (
            "If `kind` is specified, `source_exponents` and `test_exponents` should "
            "not be specified.")
        assert test_exponents is None, (
            "If `kind` is specified, `test_exponents` should not be specified.")
        assert plane is None, (
            "If `kind` is specified, `plane` should not be specified.")
        assert kind in KIND_DEFINITIONS, f'Unknown kind {kind}'
        source_exponents, test_exponents, plane = (
            KIND_DEFINITIONS[kind]['source_exponents'],
            KIND_DEFINITIONS[kind]['test_exponents'],
            KIND_DEFINITIONS[kind]['plane'])

    if exponents is not None:
        assert source_exponents is None and test_exponents is None, (
            "If exponents is specified, source_exponents and test_exponents "
            "should not be specified.")
        source_exponents = exponents[0:2]
        test_exponents = exponents[2:4]
    elif source_exponents is not None or test_exponents is not None:
        assert source_exponents is not None and test_exponents is not None, (
            "If source_exponents or test_exponents is specified, both should "
            "be specified.")
        assert exponents is None, (
            "If source_exponents or test_exponents is specified, exponents "
            "should not be specified.")

    return source_exponents, test_exponents, plane



# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import xobjects as xo
import xwakes as xw

import numpy as np
import pathlib

data_folder = pathlib.Path(__file__).parent.joinpath('../test_data').absolute()

def test_headail_table_read():
    wake_table_name = (data_folder / 'headtail_format_table_hllhc' / 
                       'HLLHC_wake_flattop_nocrab.dat')
    wake_file_columns = ['time', 'longitudinal', 'dipolar_x', 'dipolar_y',
                        'quadrupolar_x', 'quadrupolar_y', 'dipolar_xy',
                        'quadrupolar_xy', 'dipolar_yx', 'quadrupolar_yx',
                        'constant_x', 'constant_y']
    components = wake_file_columns[1:]
    wake_df = xw.read_headtail_file(wake_table_name,
                                    wake_file_columns)

    wake_data = np.loadtxt(wake_table_name)

    conversion_factor_time = -1E-9
    itime = wake_file_columns.index('time')

    dict_components = {}

    dict_components['time'] = conversion_factor_time * wake_data[:, itime]

    for i_component, component in enumerate(wake_file_columns):
        if component != 'time':
            if component == 'longitudinal':
                conversion_factor = 1E12
            else:
                conversion_factor = 1E15

            dict_components[component] = (wake_data[:, i_component] *
                                        conversion_factor)

    for component in components:
        xo.assert_allclose(dict_components[component],
                            wake_df[component].values)



# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import numpy as np
import pathlib

from xwakes.yokoya import Yokoya

data_folder = pathlib.Path(__file__).parent.joinpath('../test_data').absolute()

def test_yokoya_circular():
    yokoya_table = np.loadtxt(data_folder / 'yokoya_factors_elliptic.dat',
                              skiprows=1)[:, 1:]
    yokoya = Yokoya(shape='circular')

    for comp in yokoya.keys():
        assert np.isclose(yokoya[comp],
                          yokoya_table[0, yokoya.keys().index(comp)],
                          atol=1e-2)

def test_yokoya_flat_vertical():
    yokoya_table = np.loadtxt(data_folder / 'yokoya_factors_elliptic.dat',
                              skiprows=1)[:, 1:]
    yokoya = Yokoya(shape='flat_vertical')

    for comp in yokoya.keys():
        assert np.isclose(yokoya[comp],
                          yokoya_table[-1, yokoya.keys().index(comp)],
                          atol=1e-2)

def test_yokoya_flat_horizontal():
    yokoya_v = Yokoya(shape='flat_vertical')
    yokoya_h = Yokoya(shape='flat_horizontal')

    assert yokoya_v['dipolar_x'] == yokoya_h['dipolar_y']
    assert yokoya_v['dipolar_y'] == yokoya_h['dipolar_x']



# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import pandas as pd
import numpy as np
import pytest
from scipy.constants import c as clight

import xtrack as xt
import xpart as xp
import xwakes as xw
import xobjects as xo

@pytest.mark.parametrize('wake_type', ['dipolar', 'quadrupolar'])
@pytest.mark.parametrize('plane', ['x', 'y'])
def test_tune_shift_transverse(wake_type, plane):

    table = pd.DataFrame({'time': [0, 10],
                        f'{wake_type}_{plane}': [6e15, 6e15]})

    wake = xw.WakeFromTable(table)
    wake.configure_for_tracking(zeta_range=(-20e-2, 20e-2), num_slices=100)

    assert len(wake.components) == 1
    assert wake.components[0].plane == plane
    assert wake.components[0].source_exponents == {
        'dipolar_x': (1, 0), 'dipolar_y': (0, 1),
        'quadrupolar_x': (0, 0), 'quadrupolar_y': (0, 0)}[f'{wake_type}_{plane}']
    assert wake.components[0].test_exponents == {
        'dipolar_x': (0, 0), 'dipolar_y': (0, 0),
        'quadrupolar_x': (1, 0), 'quadrupolar_y': (0, 1)}[f'{wake_type}_{plane}']

    beta0_check = 0.3
    xo.assert_allclose(wake.components[0].function_vs_t(0.1, beta0=beta0_check, dt=1e-20), 6e15, rtol=1e-10, atol=0)
    xo.assert_allclose(wake.components[0].function_vs_t(0, beta0=beta0_check, dt=1e-20), 6e15/2, rtol=1e-10, atol=0)
    xo.assert_allclose(wake.components[0].function_vs_t(-0.1, beta0=beta0_check, dt=1e-20), 0, rtol=1e-10, atol=0)
    xo.assert_allclose(wake.components[0].function_vs_t(10 - 0.1, beta0=beta0_check, dt=1e-20), 6e15, rtol=1e-10, atol=0)
    xo.assert_allclose(wake.components[0].function_vs_t(10 + 0.1, beta0=beta0_check, dt=1e-20), 0, rtol=1e-10, atol=0)

    xo.assert_allclose(wake.components[0].function_vs_zeta(-0.1, beta0=beta0_check, dzeta=1e-20), 6e15, rtol=1e-10, atol=0)
    xo.assert_allclose(wake.components[0].function_vs_zeta(0, beta0=beta0_check, dzeta=1e-20), 6e15/2, rtol=1e-10, atol=0)
    xo.assert_allclose(wake.components[0].function_vs_zeta(0.1, beta0=beta0_check, dzeta=1e-20), 0, rtol=1e-10, atol=0)
    xo.assert_allclose(
        wake.components[0].function_vs_zeta(-beta0_check * clight * 10 + 0.1, beta0=beta0_check, dzeta=1e-20),
        6e15, rtol=1e-10, atol=0)
    xo.assert_allclose(
        wake.components[0].function_vs_zeta(-beta0_check * clight * 10 - 0.1, beta0=beta0_check, dzeta=1e-20),
        0, rtol=1e-10, atol=0)

    one_turn_map = xt.LineSegmentMap(length=1, qx=0.28, qy=0.31, qs=0.1e-3, bets=100)

    line_no_wake = xt.Line(elements=[one_turn_map])
    line_with_wake = xt.Line(elements=[one_turn_map, wake])

    line_no_wake.particle_ref = xt.Particles(p0c=2e9)
    line_with_wake.particle_ref = xt.Particles(p0c=2e9)

    line_no_wake.build_tracker()
    line_with_wake.build_tracker()

    p = xp.generate_matched_gaussian_bunch(line=line_no_wake,
                num_particles=1000, nemitt_x=1e-6, nemitt_y=1e-6, sigma_z=0.07,
                total_intensity_particles=1e11)
    p.x += 1e-3
    p.y += 1e-3

    p0 = p.copy()

    mylog = xt.Log(mean_x=lambda line, part: part.x.mean(),
                mean_y=lambda line, part: part.y.mean())

    line_no_wake.track(p0, num_turns=100,
                    log=mylog,
                    with_progress=10)
    log_no_wake = line_no_wake.log_last_track

    line_with_wake.track(p, num_turns=100,
                            log=mylog,
                            with_progress=10)
    log_with_wake = line_with_wake.log_last_track

    import nafflib as nl

    tune_no_wake = nl.get_tune(log_no_wake[f'mean_{plane}'])
    tune_with_wake = nl.get_tune(log_with_wake[f'mean_{plane}'])

    print(f'Tune without wake: {tune_no_wake}')
    print(f'Tune with wake: {tune_with_wake}')
    print(f'Tune shift: {tune_with_wake - tune_no_wake}')

    # Expect negative tune shift for positive wake
    xo.assert_allclose(tune_no_wake, {'x': 0.28, 'y': 0.31}[plane], atol=1e-6, rtol=0)
    xo.assert_allclose(tune_with_wake, tune_no_wake - 2e-3, atol=0.3e-3, rtol=0)

def test_tune_shift_longitudinal():

    table = pd.DataFrame({'time': [0, 10],
                        'longitudinal': [1e13, 1e13]})

    wake = xw.WakeFromTable(table)
    wake.configure_for_tracking(zeta_range=(-20e-2, 20e-2), num_slices=100)

    assert len(wake.components) == 1
    assert wake.components[0].plane == 'z'
    assert wake.components[0].source_exponents == (0, 0)
    assert wake.components[0].test_exponents == (0, 0)

    beta0_check = 0.3
    xo.assert_allclose(wake.components[0].function_vs_t(0.1, beta0=beta0_check, dt=1e-20), 1e13, rtol=1e-10, atol=0)
    xo.assert_allclose(wake.components[0].function_vs_t(0, beta0=beta0_check, dt=1e-20), 1e13/2, rtol=1e-10, atol=0)
    xo.assert_allclose(wake.components[0].function_vs_t(-0.1, beta0=beta0_check, dt=1e-20), 0, rtol=1e-10, atol=0)
    xo.assert_allclose(wake.components[0].function_vs_t(10 - 0.1, beta0=beta0_check, dt=1e-20), 1e13, rtol=1e-10, atol=0)
    xo.assert_allclose(wake.components[0].function_vs_t(10 + 0.1, beta0=beta0_check, dt=1e-20), 0, rtol=1e-10, atol=0)

    xo.assert_allclose(wake.components[0].function_vs_zeta(-0.1, beta0=beta0_check, dzeta=1e-20), 1e13, rtol=1e-10, atol=0)
    xo.assert_allclose(wake.components[0].function_vs_zeta(0, beta0=beta0_check, dzeta=1e-20), 1e13/2, rtol=1e-10, atol=0)
    xo.assert_allclose(wake.components[0].function_vs_zeta(0.1, beta0=beta0_check, dzeta=1e-20), 0, rtol=1e-10, atol=0)
    xo.assert_allclose(
        wake.components[0].function_vs_zeta(-beta0_check * clight * 10 + 0.1, beta0=beta0_check, dzeta=1e-20),
        1e13, rtol=1e-10, atol=0)
    xo.assert_allclose(
        wake.components[0].function_vs_zeta(-beta0_check * clight * 10 - 0.1, beta0=beta0_check, dzeta=1e-20),
        0, rtol=1e-10, atol=0)

    one_turn_map = xt.LineSegmentMap(length=1, qx=0.28, qy=0.31, qs=5e-3, bets=100)

    line_no_wake = xt.Line(elements=[one_turn_map])
    line_with_wake = xt.Line(elements=[one_turn_map, wake])

    line_no_wake.particle_ref = xt.Particles(p0c=2e9)
    line_with_wake.particle_ref = xt.Particles(p0c=2e9)

    line_no_wake.build_tracker()
    line_with_wake.build_tracker()

    p = xp.generate_matched_gaussian_bunch(line=line_no_wake,
                num_particles=1000, nemitt_x=1e-6, nemitt_y=1e-6, sigma_z=0.07,
                total_intensity_particles=1e11)
    p.zeta += 5e-3

    p0 = p.copy()

    mylog = xt.Log(mean_zeta=lambda line, part: part.zeta.mean(),
                mean_delta=lambda line, part: part.delta.mean())

    line_no_wake.track(p0, num_turns=3000,
                    log=mylog,
                    with_progress=10)
    log_no_wake = line_no_wake.log_last_track

    line_with_wake.track(p, num_turns=3000,
                            log=mylog,
                            with_progress=10)
    log_with_wake = line_with_wake.log_last_track

    import nafflib as nl

    tune_no_wake = nl.get_tune(log_no_wake[f'mean_zeta'])
    tune_with_wake = nl.get_tune(log_with_wake[f'mean_zeta']-np.mean(log_with_wake[f'mean_zeta']))

    print(f'Tune without wake: {tune_no_wake}')
    print(f'Tune with wake: {tune_with_wake}')
    print(f'Tune shift: {tune_with_wake - tune_no_wake}')

    xo.assert_allclose(tune_no_wake, 5e-3, atol=1e-6, rtol=0)
    xo.assert_allclose(tune_with_wake, tune_no_wake + 1.15e-3, atol=1e-4, rtol=0)


# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import pandas as pd
import numpy as np

def read_headtail_file(wake_file, wake_file_columns):
    valid_wake_components = ['constant_x', 'constant_y', 'dipolar_x',
                             'dipolar_y', 'dipolar_xy', 'dipolar_yx',
                             'dipole_x', 'dipole_y', 'dipole_xy',
                             'dipole_yx', 'quadrupolar_x', 'quadrupolar_y',
                             'quadrupolar_xy', 'quadrupolar_yx',
                             'quadrupole_x', 'quadrupole_y',
                             'quadrupole_xy', 'quadrupole_yx',
                             'longitudinal']

    wake_data = np.loadtxt(wake_file)
    if len(wake_file_columns) != wake_data.shape[1]:
        raise ValueError("Length of wake_file_columns list does not" +
                            " correspond to the number of columns in the" +
                            " specified wake_file. \n")
    if 'time' not in wake_file_columns:
        raise ValueError("No wake_file_column with name 'time' has" +
                            " been specified. \n")

    dict_components = {}

    conversion_factor_time = 1E-9

    itime = wake_file_columns.index('time')
    dict_components['time'] = conversion_factor_time * wake_data[:, itime]

    for i_component, component in enumerate(wake_file_columns):
        if component != 'time':
            assert component in valid_wake_components
            if component == 'longitudinal':
                conversion_factor = 1E12
            else:
                conversion_factor = 1E15

            dict_components[component] = (wake_data[:, i_component] *
                                            conversion_factor)

    df = pd.DataFrame(data=dict_components)

    return df


# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from pywit.utilities import (create_resonator_component, create_resonator_element,
                             create_many_resonators_element,
                             create_resistive_wall_single_layer_approx_component,
                             create_resistive_wall_single_layer_approx_element,
                             create_taper_RW_approx_component,
                             create_element_from_table,
                             create_interpolated_component,
                             create_component_from_arrays)

from xwakes.wit.component import ComponentSingleLayerResistiveWall
from xwakes.wit.interface_dataclasses import _IW2DInputBase
from xwakes.wit.component import KIND_DEFINITIONS

import xobjects as xo

from test_common import relative_error
from pywit.parameters import *
from pywit.interface import (FlatIW2DInput, RoundIW2DInput, Sampling,
                             component_names)
from pywit.materials import layer_from_json_material_library, copper_at_temperature

from typing import Dict
from pathlib import Path

from pytest import raises, mark, fixture
import numpy as np
from numpy import testing as npt
import pandas as pd

def test_incompatible_dictionaries():
    rs = {"z0000": 2}
    qs = {"z0000": 2}
    fs = {"x0000": 2}
    with raises(AssertionError):
        create_resonator_element(1, 1, 1, rs, qs, fs)


def test_resonator_functions():
    # Assumes transverse shunt-impedance of 10^6 Ohm/m, and for simplicity, parallel shunt-impedance of 10^6 Ohm
    # Assumes resonance frequency of 5x10^9 Hz
    # Resulting values calculated using WolframAlpha
    params = [
        (1, 1e3, 1e-8,

         0.20000000000000799999999999999999999999998719999999999948800 +
         999999.9999999999999999999983999999999999360000000000000000 * 1j,

         4.000000000000159999999999999999999999999743999999999989760 * 1e-8 +
         0.1999999999999999999999999996799999999999872000000000000000 * 1j,

         2.0790568962711056134022117523654280541098216352367107167170 * 1e-52,

         -1.64047989255734044430680584824414225044330450145887282475 * 1e-52),

        (1, 2e6, 1e-9,

         400.0000639999999999983615997378560000000067108874737418239999725 +
         999999.999999974399995904000000000104857616777215999999570503202 * 1j,

         0.160000025599999999999344639895142400000002684354989496729599989 +
         399.999999999989759998361600000000041943046710886399999828201280 * 1j,

         4.78853721755932413794457670247279476645166922134243566471030 * 1e9,

         -4.6784101513043641063808617269665441974594819336165860702670 * 1e9),

        (1, 5e9, 1e-10,

         1e6,

         1e6,

         3.08108843686609735291707294749863455067119767175475504569435 * 1e15,

         -7.501299063383624459887240816560900288553129357784921049140 * 1e15),

        (1000, 1e3, 1e-8,

         2.000000000000159999920000009599987200003711998720000793599 * 1e-7 +
         1000.000000000039999960000001599995200001663999616000322559 * 1j,

         4.00000000000031999984000001919997440000742399744000158719 * 1e-14 +
         0.000200000000000007999992000000319999040000332799923200064511 * 1j,

         -1.0543650997439856725369172254858242688504101104167514427198 * 1e9,

         2.68491822708118498041028798429290190516850878075318261084845 * 1e13),

        (1000, 2e6, 1e-9,

         0.000400000127999966719965593595166724839180685593689238045309985 +
         1000.00015999986559992729600167937403126177831194885763757974 * 1j,

         1.600000511999866879862374380666899356722742374756952181239 * 1e-7 +
         0.4000000639999462399709184006717496125047113247795430550318 * 1j,

         -1.2144732628197967915943457947230343699439427175360563689881 * 1e8,

         3.09263019467959901903826466633770980448650547056418446854141 * 1e13),

        (1000, 5e9, 1e-10,

         1e6,

         1e6,

         1.23176441001988511772523795607220037036962745466914210177191 * 1e7,

         -3.136661725760914558037302917413121144031282294807848994107 * 1e13)
    ]
    for q, f, t, z_transverse, z_parallel, w_transverse, w_parallel in params:
        transverse = create_resonator_component('x', (0, 0, 0, 0), 1e6, q, 5e9)
        parallel = create_resonator_component('z', (0, 0, 0, 0), 1e6, q, 5e9)
        assert relative_error(transverse.impedance(f), z_transverse) < REL_TOL
        assert relative_error(parallel.impedance(f), z_parallel) < REL_TOL
        assert relative_error(transverse.wake(t), w_transverse) < REL_TOL
        assert relative_error(parallel.wake(t), w_parallel) < REL_TOL


def test_resonator():
    names = ['dipx', 'dipy', 'quadx', 'quady']
    data_directory = Path(__file__).parent.joinpath('test_data').joinpath('resonator')
    params: Dict = np.load(data_directory.joinpath('dict_gen.npy'),
                           allow_pickle=True).item()

    conversion = {'dipx': ('x', (1, 0, 0, 0)),
                  'dipy': ('y', (0, 1, 0, 0)),
                  'quadx': ('x', (0, 0, 1, 0)),
                  'quady': ('y', (0, 0, 0, 1))}
    separated_components = {name: [] for name in names}
    for dim, param_sets in params.items():
        plane, exponents = conversion[dim]
        for (shunt, q, f_res) in param_sets.values():
            separated_components[dim].append(create_resonator_component(plane, exponents, shunt, q, f_res))

    freq_lin = np.arange(1e7, 1e10, 5e5)
    freq_log = np.logspace(1, 12, 625 * 11)
    freq_mesh = np.sort(np.append(freq_lin, freq_log))
    freq_mesh = freq_mesh[::100]

    time_lin = np.arange(-12e-9, 205e-9, 5e-13)
    time_log = np.logspace(-10, np.log10(3e-5), 100 * 5)
    time = np.sort(np.append(time_lin, time_log))
    time = np.delete(time, np.argwhere(np.abs(time) < 1e-20))
    time_mesh = time[time > 0]
    time_mesh = time_mesh[::100]

    for k, v in separated_components.items():
        separated_components[k] = sum(v)

    correct_impedances = np.loadtxt(data_directory.joinpath('resonator_impedances.txt'),
                                    delimiter='\t', skiprows=0)
    correct_wakes = np.loadtxt(data_directory.joinpath('resonator_wakes.txt'),
                               delimiter='\t', skiprows=0)

    correct_impedances = {'dipx': correct_impedances[:, 1] + 1j * correct_impedances[:, 2],
                          'dipy': correct_impedances[:, 3] + 1j * correct_impedances[:, 4],
                          'quadx': correct_impedances[:, 5] + 1j * correct_impedances[:, 6],
                          'quady': correct_impedances[:, 7] + 1j * correct_impedances[:, 8]}
    correct_wakes = {name: data for name, data in zip(names, [correct_wakes[:, i] for i in range(1, 5)])}

    impedances = {name: separated_components[name].impedance(freq_mesh) for name in names}
    wakes = {name: separated_components[name].wake(time_mesh) / 1e15 for name in names}

    for correct, calculated in zip(correct_impedances.values(), impedances.values()):
        np.testing.assert_allclose(correct, calculated)

    for correct, calculated in zip(correct_wakes.values(), wakes.values()):
        np.testing.assert_allclose(correct, calculated)


FREQS_LIST = [10, 20, 40, 60, 80, 100]


@mark.parametrize('test_freq', FREQS_LIST)
def test_many_resonators(test_freq):
    params_dict = {
        'z0000':
            [
                {'r': 10, 'q': 10, 'f': 50},
                {'r': 40, 'q': 100, 'f': 60}
            ],
        'x1000':
            [
                {'r': 20, 'q': 10, 'f': 40},
                {'r': 50, 'q': 100, 'f': 70}
            ],
        'y0100':
            [
                {'r': 30, 'q': 10, 'f': 70},
                {'r': 60, 'q': 100, 'f': 80}
            ]
    }

    elem = create_many_resonators_element(length=1, beta_x=1, beta_y=1, params_dict=params_dict)

    elem1 = create_resonator_element(length=1, beta_x=1, beta_y=1,
                                     rs={'z0000': params_dict['z0000'][0]['r'],
                                         'x1000': params_dict['x1000'][0]['r'],
                                         'y0100': params_dict['y0100'][0]['r']},
                                     qs={'z0000': params_dict['z0000'][0]['q'],
                                         'x1000': params_dict['x1000'][0]['q'],
                                         'y0100': params_dict['y0100'][0]['q']},
                                     fs={'z0000': params_dict['z0000'][0]['f'],
                                         'x1000': params_dict['x1000'][0]['f'],
                                         'y0100': params_dict['y0100'][0]['f']}
                                     )

    elem2 = create_resonator_element(length=1, beta_x=1, beta_y=1,
                                     rs={'z0000': params_dict['z0000'][1]['r'],
                                         'x1000': params_dict['x1000'][1]['r'],
                                         'y0100': params_dict['y0100'][1]['r']},
                                     qs={'z0000': params_dict['z0000'][1]['q'],
                                         'x1000': params_dict['x1000'][1]['q'],
                                         'y0100': params_dict['y0100'][1]['q']},
                                     fs={'z0000': params_dict['z0000'][1]['f'],
                                         'x1000': params_dict['x1000'][1]['f'],
                                         'y0100': params_dict['y0100'][1]['f']}
                                     )

    sum_elem = elem1 + elem2

    for comp in ['z0000', 'x1000', 'y0100']:
        assert elem.get_component(comp).impedance(test_freq) == sum_elem.get_component(comp).impedance(test_freq)


def test_f_roi_resonator_component():
    r = 1e9
    f_roi_level = 0.3
    resonator_component_z = create_resonator_component(plane='z', exponents=(0, 0, 0, 0), r=1e9, q=1e6, f_r=400e6,
                                                       f_roi_level=f_roi_level)
    assert np.isclose(np.real(resonator_component_z.impedance(resonator_component_z.f_rois[0][0]))/r, f_roi_level)

    resonator_component_x = create_resonator_component(plane='x', exponents=(1, 0, 0, 0), r=1e9, q=1e6, f_r=400e6,
                                                       f_roi_level=f_roi_level)
    assert np.isclose(np.real(resonator_component_x.impedance(resonator_component_x.f_rois[0][0])/r), f_roi_level)


@fixture
def flat_symmetric_input():
    return FlatIW2DInput(machine='LHC', length=1.,
            relativistic_gamma=7460.5232328, calculate_wake=False,
            f_params=Sampling(1e3,1e13,0,added=(1e4,1e9),points_per_decade=0),
            top_bottom_symmetry=True,
            top_layers=[layer_from_json_material_library(thickness=np.inf,
                                                        material_key='Mo')],
            top_half_gap=0.002)


@fixture
def flat_single_plate_input():
    return FlatIW2DInput(machine='LHC', length=1.,
            relativistic_gamma=7460.5232328, calculate_wake=False,
            f_params=Sampling(1e3,1e13,0,added=(1e5,1e9),points_per_decade=0),
            top_bottom_symmetry=False,
            top_layers=[layer_from_json_material_library(thickness=np.inf,
                                                        material_key='graphite')],
            top_half_gap=0.004,
            bottom_layers=[],
            bottom_half_gap=np.inf,
            )


@fixture
def round_single_layer_input():
    return RoundIW2DInput(machine='LHC', length=0.03,
            relativistic_gamma=479.6, calculate_wake=False,
            f_params=Sampling(1e3,1e13,0,added=(1e8),points_per_decade=0),
            layers=[copper_at_temperature(thickness=np.inf,T=293)],
            inner_layer_radius=0.02,
            yokoya_factors=(1,1,1,0,0),
            )


@mark.parametrize("component_id", ['zlong', 'zxdip', 'zydip', 'zxqua', 'zyqua'])
def test_single_plate_RW_approx_error(component_id):

    with raises(NotImplementedError):
        flat_input = FlatIW2DInput(machine='LHC', length=1.,
            relativistic_gamma=7460.5232328, calculate_wake=False,
            f_params=Sampling(1e3,1e13,0,added=(1e4,1e9),points_per_decade=0),
            top_bottom_symmetry=False,
            top_layers=[layer_from_json_material_library(thickness=np.inf,
                                                         material_key='Mo')],
            top_half_gap=0.002,
            bottom_layers=[],
            bottom_half_gap=0.025,
            )
        _ , plane, exponents = component_names[component_id]
        create_resistive_wall_single_layer_approx_component(
                plane=plane,exponents=exponents,input_data=flat_input)


@mark.parametrize("component_id", ['zlong', 'zxdip', 'zydip', 'zxqua', 'zyqua'])
def test_more_than_one_layer_RW_approx_error(component_id):

    _ , plane, exponents = component_names[component_id]

    flat_input = FlatIW2DInput(
        machine='LHC', length=1.,
        relativistic_gamma=7460.5232328, calculate_wake=False,
        f_params=Sampling(1e3,1e13,0,added=(1e4,1e9),points_per_decade=0),
        top_bottom_symmetry=True,
        top_layers=[layer_from_json_material_library(thickness=0.002,
                                                     material_key='Mo'),
                    layer_from_json_material_library(thickness=np.inf,
                                                     material_key='W')],
        top_half_gap=0.002,
        )

    with raises(NotImplementedError):
        create_resistive_wall_single_layer_approx_component(
                plane=plane,exponents=exponents,input_data=flat_input)

    with raises(NotImplementedError):
        create_taper_RW_approx_component(
                plane=plane,exponents=exponents,input_data=flat_input,
                radius_small=0.001,radius_large=0.002)

    round_input = RoundIW2DInput(
        machine='LHC', length=2.,
        relativistic_gamma=479.6, calculate_wake=False,
        f_params=Sampling(1e4,1e10,0,added=(1e12,),points_per_decade=0),
        layers=[layer_from_json_material_library(thickness=0.02,
                                                     material_key='Mo'),
                    layer_from_json_material_library(thickness=np.inf,
                                                     material_key='graphite')],
        inner_layer_radius=0.004,
        yokoya_factors=(1,1,1,0,0),
        )

    with raises(NotImplementedError):
        create_resistive_wall_single_layer_approx_component(
                plane=plane,exponents=exponents,input_data=round_input)

    with raises(NotImplementedError):
        create_taper_RW_approx_component(
                plane=plane,exponents=exponents,input_data=round_input,
                radius_small=0.01,radius_large=0.02)


@mark.parametrize("component_id", ['zlong', 'zxdip', 'zydip', 'zxqua', 'zyqua'])
def test_wrong_input_RW_approx_error(component_id):

    _ , plane, exponents = component_names[component_id]

    input_data = _IW2DInputBase(
        machine='LHC', length=1.,
        relativistic_gamma=7460.5232328, calculate_wake=False,
        f_params=Sampling(1e3,1e13,0,added=(1e4,1e9),points_per_decade=0),
        )

    with raises(NotImplementedError):
        create_resistive_wall_single_layer_approx_component(
                plane=plane,exponents=exponents,input_data=input_data)

    with raises(NotImplementedError):
        create_taper_RW_approx_component(
                plane=plane,exponents=exponents,input_data=input_data,
                radius_small=0.001,radius_large=0.002)


# expected impedance values obtained from an exact flat chamber IW2D computation
@mark.parametrize("freq, component_id, component_str, expected_Z, rtol",
                  [
                    [1e4, 'z0000', 'zlong', 2.73790435e-03+1j*3.52379262e-03, 5e-2],
                    [1e4, 'x1000', 'zxdip', 1.41444672e+06+1j*2.94098193e+06, 5e-2],
                    [1e4, 'y0100', 'zydip', 2.90873693e+06+1j*5.92025573e+06, 5e-2],
                    [1e4, 'x0010', 'zxqua',-1.41444672e+06-1j*2.94098193e+06, 5e-2],
                    [1e4, 'y0001', 'zyqua', 1.41444672e+06+1j*2.94098193e+06, 5e-2],
                    [1e9, 'z0000', 'zlong', 1.15540625e+00+1j*1.15680934e+00, 2e-3],
                    [1e9, 'x1000', 'zxdip', 1.13144575e+04+1j*1.13465281e+04, 2e-3],
                    [1e9, 'y0100', 'zydip', 2.26466174e+04+1j*2.26502884e+04, 2e-3],
                    [1e9, 'x0010', 'zxqua',-1.13144575e+04-1j*1.13465281e+04, 2e-3],
                    [1e9, 'y0001', 'zyqua', 1.13144575e+04+1j*1.13465281e+04, 2e-3],
                   ]
                  )
def test_single_layer_RW_approx_flat_sym(freq, component_id, component_str,
                                         expected_Z, rtol, flat_symmetric_input):

    elem = create_resistive_wall_single_layer_approx_element(
            input_data=flat_symmetric_input, beta_x=10,beta_y=10,
            component_ids=(component_str,),
            )
    npt.assert_allclose(elem.get_component(component_id).impedance(freq),
                        expected_Z, rtol=rtol)


# expected impedance values obtained from an exact flat chamber IW2D computation
@mark.parametrize("freq, component_id, component_str, expected_Z, rtol",
                  [
                    [1e5, 'z0000', 'zlong', 5.01728296e-02+1j*8.19175077e-02, 3e-2],
                    [1e5, 'x1000', 'zxdip', 1.73918933e+05+1j*7.32919146e+05, 3e-2],
                    [1e5, 'y0100', 'zydip', 1.73918933e+05+1j*7.32919146e+05, 3e-2],
                    [1e5, 'x0010', 'zxqua',-1.73918933e+05-1j*7.32919146e+05, 3e-2],
                    [1e5, 'y0001', 'zyqua', 1.73918933e+05+1j*7.32919146e+05, 3e-2],
                    [1e9, 'z0000', 'zlong', 9.58598752e+00+1j*9.71657980e+00, 2e-3],
                    [1e9, 'x1000', 'zxdip', 1.40485975e+04+1j*1.44908306e+04, 1e-3],
                    [1e9, 'y0100', 'zydip', 1.40485975e+04+1j*1.44908306e+04, 1e-3],
                    [1e9, 'x0010', 'zxqua',-1.40485975e+04-1j*1.44908306e+04, 1e-3],
                    [1e9, 'y0001', 'zyqua', 1.40485975e+04+1j*1.44908306e+04, 1e-3],
                   ]
                  )
def test_single_layer_RW_approx_single_plate(freq, component_id, component_str,
                                             expected_Z, rtol, flat_single_plate_input):

    elem = create_resistive_wall_single_layer_approx_element(
            input_data=flat_single_plate_input, beta_x=1,beta_y=1,
            component_ids=(component_str,),
            )
    npt.assert_allclose(elem.get_component(component_id).impedance(freq),
                        expected_Z, rtol=rtol)


# expected impedance values obtained from an exact round chamber IW2D computation
# (except xquad and yquad which are set to zero)
@mark.parametrize("freq, component_id, component_str, expected_Z, rtol",
                  [
                    [1e8, 'z0000', 'zlong', 6.18359811e-04+1j*7.73533991e-04, 1e-6],
                    [1e8, 'x1000', 'zxdip', 1.47471685e+00+1j*1.49501534e+00, 1e-6],
                    [1e8, 'y0100', 'zydip', 1.47471685e+00+1j*1.49501534e+00, 1e-6],
                    [1e8, 'x0010', 'zxqua', 0, 1e-15],
                    [1e8, 'y0001', 'zyqua', 0, 1e-15],
                   ]
                  )
def test_single_layer_RW_approx_round(freq, component_id, component_str,
                                      expected_Z, rtol, round_single_layer_input):

    elem = create_resistive_wall_single_layer_approx_element(
            input_data=round_single_layer_input, beta_x=1,beta_y=1,
            component_ids=(component_str,),
            )
    npt.assert_allclose(elem.get_component(component_id).impedance(freq),
                        expected_Z, rtol=rtol)


# test algorithm for taper RW impedance (trapz vs. step-wise integral)
@mark.parametrize("component_id", ['zlong', 'zxdip'])
@mark.parametrize("freq", [1e9, np.array([1e4, 1e8])])
def test_taper_RW_algorithm(freq, component_id, round_single_layer_input):

    _, plane, exponents = component_names[component_id]
    radius_small = 0.005
    radius_large = 0.02

    comp_taper_trapz = create_taper_RW_approx_component(
                plane, exponents, round_single_layer_input,
                radius_small=radius_small, radius_large=radius_large,
                step_size=1e-5,
                )

    step_size = 1e-6
    npts = int(np.floor(abs(radius_large-radius_small)/step_size)+1)
    radii = np.linspace(radius_small,radius_large,npts)
    z_steps = 0.

    for radius in radii:
        if component_id == 'zlong':
            z_steps += ComponentSingleLayerResistiveWall._zlong_round_single_layer_approx(freq,
                            round_single_layer_input.relativistic_gamma,
                            round_single_layer_input.layers[0], radius,
                            round_single_layer_input.length/float(npts))
        else:
            z_steps += ComponentSingleLayerResistiveWall._zdip_round_single_layer_approx(freq,
                            round_single_layer_input.relativistic_gamma,
                            round_single_layer_input.layers[0], radius,
                            round_single_layer_input.length/float(npts))

    npt.assert_allclose(comp_taper_trapz.impedance(freq), z_steps, rtol=1e-3)

def test_element_from_wake_table():
    wake_table = pd.DataFrame.from_dict({
        'time': np.linspace(0, 1, 100),
        'longitudinal': np.random.rand(100),
        'dipolar_x': np.random.rand(100),
        'dipolar_y': np.random.rand(100),
        'quadrupolar_x': np.random.rand(100),
        'quadrupolar_y': np.random.rand(100)
    })

    test_points = np.linspace(np.min(wake_table['time']),
                              np.max(wake_table['time']),
                              100)

    element = create_element_from_table(wake_table=wake_table,
                                        use_components=wake_table.columns[1:].to_list(),
                                        length=1,
                                        beta_x=1,
                                        beta_y=1,
                                        name="test element",
                                        tag="Tag",
                                        description='description')

    for component in element.components:
        for component_kind in KIND_DEFINITIONS:
            kind_definition = KIND_DEFINITIONS[component_kind]
            component_found = False
            if (kind_definition['plane'] == component.plane and
                kind_definition['test_exponents'] == component.test_exponents and
                kind_definition['source_exponents'] == component.source_exponents):
                component_found = True
                break

        assert component_found

        test_wake = np.interp(test_points, wake_table['time'], wake_table[component_kind])

        xo.assert_allclose(component.wake(test_points), test_wake, rtol=1e-3)

def test_element_from_impedance_table():
    impedance_table = pd.DataFrame.from_dict({
        'frequency': np.linspace(1e9, 1e10, 100),
        'longitudinal': np.random.rand(100),
        'dipolar_x': np.random.rand(100),
        'dipolar_y': np.random.rand(100),
        'quadrupolar_x': np.random.rand(100),
        'quadrupolar_y': np.random.rand(100)
    })

    test_points = np.linspace(np.min(impedance_table['frequency']),
                              np.max(impedance_table['frequency']),
                              100)

    element = create_element_from_table(impedance_table=impedance_table,
                                        use_components=impedance_table.columns[1:].to_list(),
                                        length=1,
                                        beta_x=1,
                                        beta_y=1,
                                        name="test element",
                                        tag="Tag",
                                        description='description')

    for component in element.components:
        for component_kind in KIND_DEFINITIONS:
            kind_definition = KIND_DEFINITIONS[component_kind]
            component_found = False
            if (kind_definition['plane'] == component.plane and
                kind_definition['test_exponents'] == component.test_exponents and
                kind_definition['source_exponents'] == component.source_exponents):
                component_found = True
                break

        assert component_found

        test_impedance = np.interp(test_points, impedance_table['frequency'],
                                   impedance_table[component_kind])

        xo.assert_allclose(component.impedance(test_points), test_impedance, rtol=1e-3)

def test_element_from_wake_and_impedance_table():
    wake_table = pd.DataFrame.from_dict({
        'time': np.linspace(0, 1, 100),
        'longitudinal': np.random.rand(100),
        'dipolar_x': np.random.rand(100),
        'dipolar_y': np.random.rand(100),
        'quadrupolar_x': np.random.rand(100),
        'quadrupolar_y': np.random.rand(100)
    })

    impedance_table = pd.DataFrame.from_dict({
        'frequency': np.linspace(1e9, 1e10, 100),
        'longitudinal': np.random.rand(100),
        'dipolar_x': np.random.rand(100),
        'dipolar_y': np.random.rand(100),
        'quadrupolar_x': np.random.rand(100),
        'quadrupolar_y': np.random.rand(100)
    })

    test_points_wake = np.linspace(np.min(wake_table['time']),
                                   np.max(wake_table['time']),
                                   100)

    test_points_impedance = np.linspace(np.min(impedance_table['frequency']),
                                        np.max(impedance_table['frequency']),
                                        100)

    element = create_element_from_table(wake_table=wake_table,
                                        impedance_table=impedance_table,
                                        use_components=wake_table.columns[1:].to_list(),
                                        length=1,
                                        beta_x=1,
                                        beta_y=1,
                                        name="test element",
                                        tag="Tag",
                                        description='description')

    for component in element.components:
        for component_kind in KIND_DEFINITIONS:
            kind_definition = KIND_DEFINITIONS[component_kind]
            component_found = False
            if (kind_definition['plane'] == component.plane and
                kind_definition['test_exponents'] == component.test_exponents and
                kind_definition['source_exponents'] == component.source_exponents):
                component_found = True
                break

        assert component_found

        assert hasattr(component, 'wake')

        if hasattr(component, 'wake'):
            test_wake = np.interp(test_points_wake, wake_table['time'],
                                  wake_table[component_kind])

            xo.assert_allclose(component.wake(test_points_wake), test_wake,
                               rtol=1e-3)

        assert hasattr(component, 'impedance')

        if hasattr(component, 'impedance'):
            test_impedance = np.interp(test_points_impedance,
                                       impedance_table['frequency'],
                                       impedance_table[component_kind])

            xo.assert_allclose(component.impedance(test_points_impedance),
                               test_impedance, rtol=1e-3)


def test_element_from_wake_table_wrong_columns():
    wake_table = pd.DataFrame.from_dict({
        'time': np.linspace(0, 1, 100),
        'longitudinal': np.random.rand(100),
        'dipolar_x': np.random.rand(100),
        'dipolar_y': np.random.rand(100),
        'quadrupolar_x': np.random.rand(100),
        'quadrupolar_y': np.random.rand(100)
    })

    with raises(ValueError):
        element = create_element_from_table(wake_table=wake_table,
                                            use_components=['wrong_column'],
                                            length=1,
                                            beta_x=1,
                                            beta_y=1,
                                            name="test element",
                                            tag="Tag",
                                            description='description')

def test_element_from_impedance_table_wrong_columns():
    impedance_table = pd.DataFrame.from_dict({
        'frequency': np.linspace(1e9, 1e10, 100),
        'longitudinal': np.random.rand(100),
        'dipolar_x': np.random.rand(100),
        'dipolar_y': np.random.rand(100),
        'quadrupolar_x': np.random.rand(100),
        'quadrupolar_y': np.random.rand(100)
    })

    with raises(ValueError):
        element = create_element_from_table(impedance_table=impedance_table,
                                            use_components=['wrong_column'],
                                            length=1,
                                            beta_x=1,
                                            beta_y=1,
                                            name="test element",
                                            tag="Tag",
                                            description='description')

def test_interpolated_component():
    interpolation_frequencies = np.linspace(0, 1, 10)
    interpolation_times = np.linspace(10, 100, 10)


    component = create_interpolated_component(
        plane='z',
        source_exponents=(0, 0),
        test_exponents=(0, 0),
        wake_input=lambda t: 2*t,
        impedance_input=lambda f: 3*f,
        interpolation_frequencies=interpolation_frequencies,
        interpolation_times=interpolation_times)

    test_points_wake = np.linspace(np.min(interpolation_times),
                                   np.max(interpolation_times),
                                   100)

    test_points_impedance = np.linspace(np.min(interpolation_frequencies),
                                        np.max(interpolation_frequencies),
                                        100)

    xo.assert_allclose(component.wake(test_points_wake),
            test_points_wake*2)
    xo.assert_allclose(component.impedance(test_points_impedance),
            test_points_impedance*3)


def test_component_from_arrays():
    interpolation_frequencies = np.linspace(0, 1, 10)
    interpolation_times = np.linspace(10, 100, 10)

    impedance_samples = interpolation_frequencies*2
    wake_samples = interpolation_times*3

    component = create_component_from_arrays(
        plane='z',
        source_exponents=(0, 0),
        test_exponents=(0, 0),
        wake_samples=wake_samples,
        impedance_samples=impedance_samples,
        interpolation_frequencies=interpolation_frequencies,
        interpolation_times=interpolation_times)

    test_points_wake = np.linspace(np.min(interpolation_times),
                                   np.max(interpolation_times),
                                   100)

    test_points_impedance = np.linspace(np.min(interpolation_frequencies),
                                        np.max(interpolation_frequencies),
                                        100)

    xo.assert_allclose(component.impedance(test_points_impedance),
            test_points_impedance*2)
    xo.assert_allclose(component.wake(test_points_wake),
            test_points_wake*3)
